# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from spinnman.processes import AbstractMultiConnectionProcess
from spinn_front_end_common.utilities.utility_objs.\
    extra_monitor_scp_messages import (
        LoadSystemMCRoutesMessage)


class LoadSystemMCRoutesProcess(AbstractMultiConnectionProcess):
    """ How to send messages to load the configured system multicast routing\
        tables (and save the application routing tables).
    """

    def load_system_mc_routes(self, core_subsets):
        """
        :param ~spinn_machine.CoreSubsets core_subsets:
            sets of cores to send command to
        """
        for core_subset in core_subsets.core_subsets:
            for processor_id in core_subset.processor_ids:
                self._send_request(LoadSystemMCRoutesMessage(
                    core_subset.x, core_subset.y, processor_id))
        self._finish()
        self.check_for_error()



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os
from spinn_utilities.log import FormatAdapter
from pacman.model.graphs.common import EdgeTrafficType
from spinn_front_end_common.utilities.sqlite_db import SQLiteDB
from spinn_front_end_common.abstract_models import (
    AbstractProvidesKeyToAtomMapping,
    AbstractSupportsDatabaseInjection)

logger = FormatAdapter(logging.getLogger(__name__))
DB_NAME = "input_output_database.db"
INIT_SQL = "db.sql"


def _extract_int(x):
    return None if x is None else int(x)


class DatabaseWriter(SQLiteDB):
    """ The interface for the database system for main front ends.\
        Any special tables needed from a front end should be done\
        by sub classes of this interface.
    """

    __slots__ = [
        # the path of the database
        "_database_path",

        # the identifier for the SpiNNaker machine
        "_machine_id",

        # Mappings used to accelerate inserts
        "__machine_to_id", "__vertex_to_id", "__edge_to_id"
    ]

    def __init__(self, database_directory):
        """
        :param str database_directory: Where the database will be written
        """
        self._database_path = os.path.join(database_directory, DB_NAME)
        init_sql_path = os.path.join(os.path.dirname(__file__), INIT_SQL)

        # delete any old database
        if os.path.isfile(self._database_path):
            os.remove(self._database_path)

        super().__init__(self._database_path, ddl_file=init_sql_path)
        self.__machine_to_id = dict()
        self.__vertex_to_id = dict()
        self.__edge_to_id = dict()

        # set up checks
        self._machine_id = 0

    @staticmethod
    def auto_detect_database(machine_graph):
        """ Auto detects if there is a need to activate the database system

        :param ~pacman.model.graphs.machine.MachineGraph machine_graph:
            the machine graph of the application problem space.
        :return: whether the database is needed for the application
        :rtype: bool
        """
        return any(isinstance(vertex, AbstractSupportsDatabaseInjection)
                   and vertex.is_in_injection_mode
                   for vertex in machine_graph.vertices)

    @property
    def database_path(self):
        """
        :rtype: str
        """
        return self._database_path

    def __insert(self, cur, sql, *args):
        """
        :param ~sqlite3.Cursor cur:
        :param str sql:
        :rtype: int
        """
        try:
            cur.execute(sql, args)
            return cur.lastrowid
        except Exception:
            logger.exception("problem with insertion; argument types are {}",
                             str(map(type, args)))
            raise

    def add_machine_objects(self, machine):
        """ Store the machine object into the database

        :param ~spinn_machine.Machine machine: the machine object.
        """
        with self.transaction() as cur:
            self.__machine_to_id[machine] = self._machine_id = self.__insert(
                cur,
                """
                INSERT INTO Machine_layout(
                    x_dimension, y_dimension)
                VALUES(?, ?)
                """, machine.width, machine.height)
            cur.executemany(
                """
                INSERT INTO Machine_chip(
                    no_processors, chip_x, chip_y, machine_id)
                VALUES (?, ?, ?, ?)
                """, (
                    (chip.n_processors, chip.x, chip.y, self._machine_id)
                    for chip in machine.chips if chip.virtual))
            cur.executemany(
                """
                INSERT INTO Machine_chip(
                    no_processors, chip_x, chip_y, machine_id,
                    ip_address, nearest_ethernet_x, nearest_ethernet_y)
                VALUES (?, ?, ?, ?, ?, ?, ?)
                """, (
                    (chip.n_processors, chip.x, chip.y, self._machine_id,
                     chip.ip_address,
                     chip.nearest_ethernet_x, chip.nearest_ethernet_y)
                    for chip in machine.chips if not chip.virtual))
            cur.executemany(
                """
                INSERT INTO Processor(
                    chip_x, chip_y, machine_id, available_DTCM,
                    available_CPU, physical_id)
                VALUES(?, ?, ?, ?, ?, ?)
                """, (
                    (chip.x, chip.y, self._machine_id, proc.dtcm_available,
                     proc.cpu_cycles_available, proc.processor_id)
                    for chip in machine.chips
                    for proc in chip.processors))

    def add_application_vertices(self, application_graph):
        """ Stores the main application graph description (vertices, edges).

        :param application_graph: The graph to add from
        :type application_graph:
            ~pacman.model.graphs.application.ApplicationGraph
        """
        with self.transaction() as cur:
            # add vertices
            for vertex in application_graph.vertices:
                self.__vertex_to_id[vertex] = self.__insert(
                    cur,
                    """
                    INSERT INTO Application_vertices(
                        vertex_label, vertex_class, no_atoms,
                        max_atom_constrant)
                    VALUES(?, ?, ?, ?)
                    """,
                    vertex.label, vertex.__class__.__name__, vertex.n_atoms,
                    vertex.get_max_atoms_per_core())

            # add edges
            for edge in application_graph.edges:
                self.__edge_to_id[edge] = self.__insert(
                    cur,
                    """
                    INSERT INTO Application_edges (
                        pre_vertex, post_vertex, edge_label, edge_class)
                    VALUES(?, ?, ?, ?)
                    """,
                    self.__vertex_to_id[edge.pre_vertex],
                    self.__vertex_to_id[edge.post_vertex],
                    edge.label, edge.__class__.__name__)

            # update graph
            cur.executemany(
                """
                INSERT INTO Application_graph (
                    vertex_id, edge_id)
                VALUES(?, ?)
                """, (
                    (self.__vertex_to_id[vertex], self.__edge_to_id[edge])
                    for vertex in application_graph.vertices
                    for edge in application_graph.get_edges_starting_at_vertex(
                        vertex)))

    def add_system_params(self, time_scale_factor, machine_time_step, runtime):
        """ Write system params into the database

        :param int time_scale_factor: the time scale factor used in timing
        :param int machine_time_step: the machine time step used in timing
        :param int runtime: the amount of time the application is to run for
        """
        with self.transaction() as cur:
            cur.executemany(
                """
                INSERT INTO configuration_parameters (
                    parameter_id, value)
                VALUES (?, ?)
                """, [
                    ("machine_time_step", machine_time_step),
                    ("time_scale_factor", time_scale_factor),
                    ("infinite_run", str(runtime is None)),
                    ("runtime", -1 if runtime is None else runtime)])

    def add_vertices(self, machine_graph, data_n_timesteps, application_graph):
        """ Add the machine graph into the database.

        :param ~pacman.model.graphs.machine.MachineGraph machine_graph:
            The machine graph object
        :param int data_n_timesteps:
            The number of timesteps for which data space will been reserved
        :param application_graph: The application graph object
        :type application_graph:
            ~pacman.model.graphs.application.ApplicationGraph
        """
        with self.transaction() as cur:
            for vertex in machine_graph.vertices:
                req = vertex.resources_required
                self.__vertex_to_id[vertex] = self.__insert(
                    cur,
                    """
                    INSERT INTO Machine_vertices (
                        label, class, cpu_used, sdram_used, dtcm_used)
                    VALUES(?, ?, ?, ?, ?)
                    """,
                    str(vertex.label), vertex.__class__.__name__,
                    _extract_int(req.cpu_cycles.get_value()),
                    _extract_int(req.sdram.get_total_sdram(data_n_timesteps)),
                    _extract_int(req.dtcm.get_value()))

            # add machine edges
            for edge in machine_graph.edges:
                self.__edge_to_id[edge] = self.__insert(
                    cur,
                    """
                    INSERT INTO Machine_edges (
                        pre_vertex, post_vertex, label, class)
                    VALUES(?, ?, ?, ?)
                    """,
                    self.__vertex_to_id[edge.pre_vertex],
                    self.__vertex_to_id[edge.post_vertex],
                    edge.label, edge.__class__.__name__)

            # add to machine graph
            cur.executemany(
                """
                INSERT INTO Machine_graph (
                    vertex_id, edge_id)
                VALUES(?, ?)
                """, (
                    (self.__vertex_to_id[vertex], self.__edge_to_id[edge])
                    for vertex in machine_graph.vertices
                    for edge in machine_graph.get_edges_starting_at_vertex(
                        vertex)))

            if application_graph is not None:
                cur.executemany(
                    """
                    INSERT INTO graph_mapper_vertex (
                        application_vertex_id, machine_vertex_id, lo_atom,
                        hi_atom)
                    VALUES(?, ?, ?, ?)
                    """, (
                        (self.__vertex_to_id[vertex.app_vertex],
                         self.__vertex_to_id[vertex],
                         vertex.vertex_slice.lo_atom,
                         vertex.vertex_slice.hi_atom)
                        for vertex in machine_graph.vertices))

                # add graph_mapper edges
                cur.executemany(
                    """
                    INSERT INTO graph_mapper_edges (
                        application_edge_id, machine_edge_id)
                    VALUES(?, ?)
                    """, (
                        (self.__edge_to_id[edge.app_edge],
                         self.__edge_to_id[edge])
                        for edge in machine_graph.edges))

    def add_placements(self, placements):
        """ Adds the placements objects into the database

        :param ~pacman.model.placements.Placements placements:
            the placements object
        """
        with self.transaction() as cur:
            # add records
            cur.executemany(
                """
                INSERT INTO Placements(
                    vertex_id, chip_x, chip_y, chip_p, machine_id)
                VALUES(?, ?, ?, ?, ?)
                """, (
                    (self.__vertex_to_id[placement.vertex],
                     placement.x, placement.y, placement.p, self._machine_id)
                    for placement in placements.placements))

    def add_routing_infos(self, routing_infos, machine_graph):
        """ Adds the routing information (key masks etc) into the database

        :param ~pacman.model.routing_info.RoutingInfo routing_infos:
            the routing information object
        :param ~pacman.model.graphs.machine.MachineGraph machine_graph:
            the machine graph object
        """
        # Filter just the MULTICAST partitions first
        partitions_and_routing_info = (
            (partition, routing_infos.get_routing_info_from_partition(
                partition))
            for partition in machine_graph.outgoing_edge_partitions
            if partition.traffic_type == EdgeTrafficType.MULTICAST)
        with self.transaction() as cur:
            cur.executemany(
                """
                INSERT INTO Routing_info(
                    edge_id, "key", mask)
                VALUES(?, ?, ?)
                """, (
                    (self.__edge_to_id[edge], key_mask.key, key_mask.mask)
                    for partition, rinfo in partitions_and_routing_info
                    for edge in partition.edges
                    for key_mask in rinfo.keys_and_masks))

    def add_routing_tables(self, routing_tables):
        """ Adds the routing tables into the database

        :param routing_tables: the routing tables object
        :type routing_tables:
            ~pacman.model.routing_tables.MulticastRoutingTables
        """
        with self.transaction() as cur:
            cur.executemany(
                """
                INSERT INTO Routing_table(
                    chip_x, chip_y, position, key_combo, mask, route)
                VALUES(?, ?, ?, ?, ?, ?)
                """, (
                    (routing_table.x, routing_table.y, counter,
                     entry.routing_entry_key, entry.mask,
                     entry.spinnaker_route)
                    for routing_table in routing_tables.routing_tables
                    for counter, entry in
                    enumerate(routing_table.multicast_routing_entries)))

    def add_tags(self, machine_graph, tags):
        """ Adds the tags into the database

        :param ~pacman.model.graphs.machine.MachineGraph machine_graph:
            the machine graph object
        :param ~pacman.model.tags.Tags tags: the tags object
        """
        with self.transaction() as cur:
            for vertex in machine_graph.vertices:
                v_id = self.__vertex_to_id[vertex]
                cur.executemany(
                    """
                    INSERT INTO IP_tags(
                        vertex_id, tag, board_address, ip_address, port,
                        strip_sdp)
                    VALUES (?, ?, ?, ?, ?, ?)
                    """, (
                        (v_id, ipt.tag, ipt.board_address, ipt.ip_address,
                         ipt.port or 0, 1 if ipt.strip_sdp else 0)
                        for ipt in tags.get_ip_tags_for_vertex(vertex) or []))
                cur.executemany(
                    """
                    INSERT INTO Reverse_IP_tags(
                        vertex_id, tag, board_address, port)
                    VALUES (?, ?, ?, ?)
                    """, (
                        (v_id, ript.tag, ript.board_address, ript.port or 0)
                        for ript in tags.get_reverse_ip_tags_for_vertex(
                            vertex) or ()))

    def create_atom_to_event_id_mapping(
            self, application_graph, machine_graph, routing_infos):
        """
        :param application_graph:
        :type application_graph:
            ~pacman.model.graphs.application.ApplicationGraph
        :param ~pacman.model.graphs.machine.MachineGraph machine_graph:
        :param ~pacman.model.routing_info.RoutingInfo routing_infos:
        """
        if application_graph is not None and application_graph.n_vertices:
            # We will be asking application vertices for key/atom mappings
            vertices_and_partitions = (
                (vertex.app_vertex, partition)
                for vertex in machine_graph.vertices
                for partition in machine_graph.
                get_outgoing_edge_partitions_starting_at_vertex(vertex))
        else:
            # We will be asking machine vertices for key/atom mappings
            vertices_and_partitions = (
                (vertex, partition)
                for vertex in machine_graph.vertices
                for partition in machine_graph.
                get_outgoing_edge_partitions_starting_at_vertex(vertex))

        with self.transaction() as cur:
            cur.executemany(
                """
                INSERT INTO event_to_atom_mapping(
                    vertex_id, event_id, atom_id)
                VALUES (?, ?, ?)
                """, (
                    (self.__vertex_to_id[vtx], int(key), int(a_id))
                    for vtx, prtn in vertices_and_partitions
                    if isinstance(vtx, AbstractProvidesKeyToAtomMapping)
                    for a_id, key in vtx.routing_key_partition_atom_mapping(
                        routing_infos.get_routing_info_from_partition(prtn),
                        prtn)))



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from spinn_utilities.overrides import overrides
from pacman.model.partitioner_interfaces import LegacyPartitionerAPI
from pacman.executor.injection_decorator import inject_items
from pacman.model.graphs.application import ApplicationVertex
from .chip_power_monitor_machine_vertex import ChipPowerMonitorMachineVertex


class ChipPowerMonitor(ApplicationVertex, LegacyPartitionerAPI):
    """ Represents idle time recording code in a application graph.
    """
    __slots__ = ["_n_samples_per_recording", "_sampling_frequency"]

    def __init__(
            self, label, n_samples_per_recording, sampling_frequency,
            constraints=None):
        """
        :param str label: vertex label
        :param constraints: constraints for the vertex
        :type constraints:
            iterable(~pacman.model.constraints.AbstractConstraint)
        :param int n_samples_per_recording:
            how many samples to take before recording to SDRAM the total
        :param int sampling_frequency: how many microseconds between sampling
        """
        super().__init__(label, constraints, 1)
        self._n_samples_per_recording = n_samples_per_recording
        self._sampling_frequency = sampling_frequency

    @property
    @overrides(LegacyPartitionerAPI.n_atoms)
    def n_atoms(self):
        return 1

    @overrides(LegacyPartitionerAPI.create_machine_vertex)
    def create_machine_vertex(
            self,
            vertex_slice, resources_required,  # @UnusedVariable
            label=None, constraints=None):
        machine_vertex = ChipPowerMonitorMachineVertex(
            constraints=constraints, label=label,
            n_samples_per_recording=self._n_samples_per_recording,
            sampling_frequency=self._sampling_frequency,  app_vertex=self)
        if vertex_slice:
            assert (vertex_slice == machine_vertex.vertex_slice)
        if resources_required:
            assert (resources_required ==
                    machine_vertex.resources_required)
        return machine_vertex

    @inject_items({
        "machine_time_step": "MachineTimeStep",
        "time_scale_factor": "TimeScaleFactor"})
    @overrides(LegacyPartitionerAPI.get_resources_used_by_atoms,
               additional_arguments={"machine_time_step", "time_scale_factor"})
    def get_resources_used_by_atoms(
            self, vertex_slice,  # @UnusedVariable
            machine_time_step, time_scale_factor):
        """
        :param int machine_time_step:
        :param int time_scale_factor:
        """
        # pylint: disable=arguments-differ
        return ChipPowerMonitorMachineVertex.get_resources(
            machine_time_step, time_scale_factor,
            self._n_samples_per_recording, self._sampling_frequency)



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .abstract_database import AbstractDatabase
from .buffered_receiving_data import BufferedReceivingData
from .buffered_sending_region import BufferedSendingRegion
from .buffers_sent_deque import BuffersSentDeque
from .sqllite_database import SqlLiteDatabase

__all__ = ["AbstractDatabase", "BufferedReceivingData",
           "BufferedSendingRegion", "BuffersSentDeque", "SqlLiteDatabase"]



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from enum import Enum
from data_specification.constants import APP_PTR_TABLE_BYTE_SIZE

# conversion from words to bytes
BYTES_PER_WORD = 4
BYTES_PER_3_WORDS = 12
BYTES_PER_SHORT = 2
BYTES_PER_KB = 1024

LIVE_GATHERER_CORE_APPLICATION_ID = 0xAC0
COMMAND_SENDER_CORE_APPLICATION_ID = 0xAC6
SPIKE_INJECTOR_CORE_APPLICATION_ID = 0xAC9

# how many bits there are in a word
BITS_PER_WORD = 32.0

#: start of where SDRAM starts (either unbuffered or buffered)
SDRAM_BASE_ADDR = 0x70000000

#: the ITCM max safe limit for a binary
MAX_SAFE_BINARY_SIZE = 32 * BYTES_PER_KB

#: the ITCM max limit for a binary
MAX_POSSIBLE_BINARY_SIZE = 33 * BYTES_PER_KB

# converts between micro and milli seconds
MICRO_TO_MILLISECOND_CONVERSION = 1000.0
MICRO_TO_SECOND_CONVERSION = 1000000.0  # (1e6)

#: max size expected to be used by the reverse ip_tag multicast source
#: during buffered operations.
MAX_SIZE_OF_BUFFERED_REGION_ON_CHIP = 1 * 1024 * BYTES_PER_KB

#: The default size of a recording buffer before receive request is sent
DEFAULT_BUFFER_SIZE_BEFORE_RECEIVE = 16 * BYTES_PER_KB

#: The number of bytes used by SARK per memory allocation
SARK_PER_MALLOC_SDRAM_USAGE = 2 * BYTES_PER_WORD

#: The number of words in the AbstractDataSpecable basic setup information.
#: This is the amount required by the pointer table plus a SARK allocation.
DATA_SPECABLE_BASIC_SETUP_INFO_N_BYTES = (
    APP_PTR_TABLE_BYTE_SIZE + SARK_PER_MALLOC_SDRAM_USAGE)

#: The number of bytes used by the simulation interface.
#: This is one word for the machine_time_step, one for the SDP port, and one
#: for the application hash.
SIMULATION_N_BYTES = 3 * BYTES_PER_WORD

#: the number of bytes used by the multicast data speed up interface
# 4 for the first key used by multicast protocol
MULTICAST_SPEEDUP_N_BYTES = BYTES_PER_WORD

#: The number of bytes used by the DSG and simulation interfaces
SYSTEM_BYTES_REQUIREMENT = (
    DATA_SPECABLE_BASIC_SETUP_INFO_N_BYTES + SIMULATION_N_BYTES)

#: Database file path maximum length for database notification messages.
#: Note that this is *not* sent to SpiNNaker and so is not subject to the
#: usual SDP limit.
MAX_DATABASE_PATH_LENGTH = 50000

#: size of the on-chip DSE data structure required, in bytes
DSE_DATA_STRUCT_SIZE = 4 * BYTES_PER_WORD


class SDP_RUNNING_MESSAGE_CODES(Enum):
    SDP_STOP_ID_CODE = 6
    SDP_NEW_RUNTIME_ID_CODE = 7
    SDP_UPDATE_PROVENCE_REGION_AND_EXIT = 8
    SDP_CLEAR_IOBUF_CODE = 9


class SDP_PORTS(Enum):
    """SDP port handling output buffering data streaming"""

    # command port for the buffered in functionality
    INPUT_BUFFERING_SDP_PORT = 1
    # command port for the buffered out functionality
    OUTPUT_BUFFERING_SDP_PORT = 2
    # command port for resetting runtime etc
    RUNNING_COMMAND_SDP_PORT = 3
    # extra monitor core reinjection functionality
    EXTRA_MONITOR_CORE_REINJECTION = 4
    # extra monitor core data transfer functionality
    EXTRA_MONITOR_CORE_DATA_SPEED_UP = 5
    # extra monitor core data in speed up functionality
    EXTRA_MONITOR_CORE_DATA_IN_SPEED_UP = 6


# output buffering operations
class BUFFERING_OPERATIONS(Enum):
    """A listing of what SpiNNaker specific EIEIO commands there are."""

    # Database handshake with external program
    BUFFER_READ = 0
    # Host confirming data being read form SpiNNaker memory
    BUFFER_WRITE = 1


#: partition IDs preallocated to functionality
PARTITION_ID_FOR_MULTICAST_DATA_SPEED_UP = "DATA_SPEED_UP_ROAD"

#: The default local port that the toolchain listens on for the notification
#: protocol.
NOTIFY_PORT = 19999

#: The number of clock cycles per micro-second (at 200Mhz)
CLOCKS_PER_US = 200

PROVENANCE_DB = "provenance.sqlite3"



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
from spinn_utilities.progress_bar import ProgressBar


class FixedRouteFromMachineReport(object):
    """ Generate a report of the fixed routes from the machine.
    """

    def __call__(self, transceiver, machine, report_default_directory,
                 app_id):
        """ Writes the fixed routes from the machine

        :param ~spinnman.transceiver.Transceiver transceiver: spinnMan instance
        :param ~spinn_machine.Machine machine: SpiNNMachine instance
        :param str report_default_directory: folder where reports reside
        :param int app_id: the application ID the fixed routes were loaded with
        """
        file_name = os.path.join(
            report_default_directory, "fixed_route_routers")

        with open(file_name, "w") as output:
            self._write_fixed_routers(output, transceiver, machine, app_id)

    def _write_fixed_routers(self, f, txrx, machine, app_id):
        """ How to actually describe the fixed routes

        :param f: Where we are writing
        :param ~spinnman.transceiver.Transceiver txrx:
        :param ~spinn_machine.Machine machine:
        :param int app_id: Which application is running on the machine
        :rtype: None
        """
        progress = ProgressBar(machine.n_chips, "Writing fixed route report")
        f.write(" x    y       route         [cores][links]\n")
        for chip in progress.over(machine.chips):
            if not chip.virtual:
                fixed_route = txrx.read_fixed_route(chip.x, chip.y, app_id)
                f.write("{: <3s}:{: <3s} contains route {: <10s} {}\n".format(
                    str(chip.x), str(chip.y),
                    self._reduce_route_value(
                        fixed_route.processor_ids, fixed_route.link_ids),
                    self._expand_route_value(
                        fixed_route.processor_ids, fixed_route.link_ids)))

    @staticmethod
    def _reduce_route_value(processors_ids, link_ids):
        value = 0
        for link in link_ids:
            value += 1 << link
        for processor in processors_ids:
            value += 1 << (processor + 6)
        return str(value)

    @staticmethod
    def _expand_route_value(processors_ids, link_ids):
        """ Convert a 32-bit route word into a string which lists the target\
            cores and links.
        """

        # Convert processor targets to readable values:
        route_string = "["
        separator = ""
        for processor in processors_ids:
            route_string += separator + str(processor)
            separator = ", "

        route_string += "] ["

        # Convert link targets to readable values:
        link_labels = {0: 'E', 1: 'NE', 2: 'N', 3: 'W', 4: 'SW', 5: 'S'}

        separator = ""
        for link in link_ids:
            route_string += separator + link_labels[link]
            separator = ", "
        route_string += "]"
        return route_string



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys

from pacman.model.partitioner_interfaces import LegacyPartitionerAPI
from spinn_utilities.overrides import overrides
from pacman.model.graphs.application import ApplicationVertex
from pacman.model.resources import (
    CPUCyclesPerTickResource, DTCMResource, ResourceContainer,
    ReverseIPtagResource)
from pacman.model.constraints.placer_constraints import BoardConstraint
from spinn_front_end_common.abstract_models import (
    AbstractProvidesOutgoingPartitionConstraints)
from spinn_front_end_common.abstract_models.impl import (
    ProvidesKeyToAtomMappingImpl)
from spinn_front_end_common.utilities.constants import SDP_PORTS
from .reverse_ip_tag_multicast_source_machine_vertex import (
    ReverseIPTagMulticastSourceMachineVertex)
from spinn_front_end_common.utilities.exceptions import ConfigurationException
from spinn_front_end_common.utilities import globals_variables


class ReverseIpTagMultiCastSource(
        ApplicationVertex, LegacyPartitionerAPI,
        AbstractProvidesOutgoingPartitionConstraints,
        ProvidesKeyToAtomMappingImpl):
    """ A model which will allow events to be injected into a SpiNNaker\
        machine and converted into multicast packets.
    """

    def __init__(
            self, n_keys, label=None, constraints=None,
            max_atoms_per_core=sys.maxsize,

            # General parameters
            board_address=None,

            # Live input parameters
            receive_port=None,
            receive_sdp_port=SDP_PORTS.INPUT_BUFFERING_SDP_PORT.value,
            receive_tag=None,
            receive_rate=10,

            # Key parameters
            virtual_key=None, prefix=None,
            prefix_type=None, check_keys=False,

            # Send buffer parameters
            send_buffer_times=None,
            send_buffer_partition_id=None,

            # Extra flag for input without a reserved port
            reserve_reverse_ip_tag=False,

            # Flag to indicate that data will be received to inject
            enable_injection=False,

            # splitter object
            splitter=None):
        """
        :param int n_keys:
            The number of keys to be sent via this multicast source
        :param str label: The label of this vertex
        :param constraints: Any initial constraints to this vertex
        :type constraints:
            iterable(~pacman.model.constraints.AbstractConstraint)
        :param int max_atoms_per_core:
        :param board_address: The IP address of the board on which to place
            this vertex if receiving data, either buffered or live (by
            default, any board is chosen)
        :type board_address: str or None
        :param receive_port: The port on the board that will listen for
            incoming event packets (default is to disable this feature; set a
            value to enable it)
        :type receive_port: int or None
        :param int receive_sdp_port:
            The SDP port to listen on for incoming event packets
            (defaults to 1)
        :param ~spinn_machine.tags.IPTag receive_tag:
            The IP tag to use for receiving live events
            (uses any by default)
        :param float receive_rate:
            The estimated rate of packets that will be sent by this source
        :param int virtual_key:
            The base multicast key to send received events with
            (assigned automatically by default)
        :param int prefix:
            The prefix to "or" with generated multicast keys
            (default is no prefix)
        :param ~spinnman.messages.eieio.EIEIOPrefix prefix_type:
            Whether the prefix should apply to the upper or lower half of the
            multicast keys (default is upper half)
        :param bool check_keys:
            True if the keys of received events should be verified before
            sending (default False)
        :param send_buffer_times: An array of arrays of times at which keys
            should be sent (one array for each key, default disabled)
        :type send_buffer_times:
            ~numpy.ndarray(~numpy.ndarray(numpy.int32)) or
            list(~numpy.ndarray(~numpy.int32)) or None
        :param send_buffer_partition_id: The ID of the partition containing
            the edges down which the events are to be sent
        :type send_buffer_partition_id: str or None
        :param bool reserve_reverse_ip_tag:
            Extra flag for input without a reserved port
        :param bool enable_injection:
            Flag to indicate that data will be received to inject
        :param splitter: the splitter object needed for this vertex
        :type splitter: None or AbstractSplitterCommon
        """
        # pylint: disable=too-many-arguments, too-many-locals
        super().__init__(
            label, constraints, max_atoms_per_core, splitter=splitter)

        # basic items
        self._n_atoms = self.round_n_atoms(n_keys, "n_keys")

        # Store the parameters for EIEIO
        self._board_address = board_address
        self._receive_port = receive_port
        self._receive_sdp_port = receive_sdp_port
        self._receive_tag = receive_tag
        self._receive_rate = receive_rate
        self._virtual_key = virtual_key
        self._prefix = prefix
        self._prefix_type = prefix_type
        self._check_keys = check_keys

        self._reverse_iptags = None
        if receive_port is not None or reserve_reverse_ip_tag:
            self._reverse_iptags = [ReverseIPtagResource(
                port=receive_port, sdp_port=receive_sdp_port,
                tag=receive_tag)]
            if board_address is not None:
                self.add_constraint(BoardConstraint(board_address))

        # Store the send buffering details
        self._send_buffer_times = self._validate_send_buffer_times(
            send_buffer_times)
        self._send_buffer_partition_id = send_buffer_partition_id

        # Store the buffering details
        self._reserve_reverse_ip_tag = reserve_reverse_ip_tag
        self._enable_injection = enable_injection

        # Store recording parameters
        self._is_recording = False

    def _validate_send_buffer_times(self, send_buffer_times):
        if send_buffer_times is None:
            return None
        if len(send_buffer_times) and hasattr(send_buffer_times[0], "__len__"):
            if len(send_buffer_times) != self._n_atoms:
                raise ConfigurationException(
                    "The array or arrays of times {} does not have the "
                    "expected length of {}".format(
                        send_buffer_times, self._n_atoms))
        return send_buffer_times

    @property
    @overrides(LegacyPartitionerAPI.n_atoms)
    def n_atoms(self):
        return self._n_atoms

    @overrides(LegacyPartitionerAPI.get_resources_used_by_atoms)
    def get_resources_used_by_atoms(self, vertex_slice):
        send_buffer_times = self._send_buffer_times
        if send_buffer_times is not None and len(send_buffer_times):
            if hasattr(send_buffer_times[0], "__len__"):
                send_buffer_times = send_buffer_times[
                    vertex_slice.lo_atom:vertex_slice.hi_atom + 1]
                # Check the buffer times on the slice are not empty
                n_buffer_times = 0
                for i in send_buffer_times:
                    if hasattr(i, "__len__"):
                        n_buffer_times += len(i)
                    else:
                        # assuming this must be a single integer
                        n_buffer_times += 1
                if n_buffer_times == 0:
                    send_buffer_times = None

        sim = globals_variables.get_simulator()
        container = ResourceContainer(
            sdram=ReverseIPTagMulticastSourceMachineVertex.get_sdram_usage(
                send_buffer_times, self._is_recording, sim.machine_time_step,
                self._receive_rate, vertex_slice.n_atoms),
            dtcm=DTCMResource(
                ReverseIPTagMulticastSourceMachineVertex.get_dtcm_usage()),
            cpu_cycles=CPUCyclesPerTickResource(
                ReverseIPTagMulticastSourceMachineVertex.get_cpu_usage()),
            reverse_iptags=self._reverse_iptags)
        return container

    @property
    def send_buffer_times(self):
        """ When messages will be sent.

        :rtype: ~numpy.ndarray(~numpy.ndarray(numpy.int32)) or
            list(~numpy.ndarray(~numpy.int32)) or None
        """
        return self._send_buffer_times

    @send_buffer_times.setter
    def send_buffer_times(self, send_buffer_times):
        self._send_buffer_times = send_buffer_times
        for vertex in self.machine_vertices:
            vertex_slice = vertex.vertex_slice
            send_buffer_times_to_set = self._send_buffer_times
            # pylint: disable=len-as-condition
            if (self._send_buffer_times is not None and
                    len(self._send_buffer_times)):
                if hasattr(self._send_buffer_times[0], "__len__"):
                    send_buffer_times_to_set = self._send_buffer_times[
                        vertex_slice.lo_atom:vertex_slice.hi_atom + 1]
            vertex.send_buffer_times = send_buffer_times_to_set

    def enable_recording(self, new_state=True):
        self._is_recording = new_state

    @overrides(AbstractProvidesOutgoingPartitionConstraints.
               get_outgoing_partition_constraints)
    def get_outgoing_partition_constraints(self, partition):
        return partition.pre_vertex.get_outgoing_partition_constraints(
            partition)

    @overrides(LegacyPartitionerAPI.create_machine_vertex)
    def create_machine_vertex(
            self, vertex_slice,
            resources_required,  # @UnusedVariable
            label=None, constraints=None):
        send_buffer_times = self._send_buffer_times
        if send_buffer_times is not None and len(send_buffer_times):
            if hasattr(send_buffer_times[0], "__len__"):
                send_buffer_times = send_buffer_times[
                    vertex_slice.lo_atom:vertex_slice.hi_atom + 1]
        machine_vertex = ReverseIPTagMulticastSourceMachineVertex(
            vertex_slice=vertex_slice,
            label=label, constraints=constraints, app_vertex=self,
            board_address=self._board_address,
            receive_port=self._receive_port,
            receive_sdp_port=self._receive_sdp_port,
            receive_tag=self._receive_tag,
            receive_rate=self._receive_rate,
            virtual_key=self._virtual_key, prefix=self._prefix,
            prefix_type=self._prefix_type, check_keys=self._check_keys,
            send_buffer_times=send_buffer_times,
            send_buffer_partition_id=self._send_buffer_partition_id,
            reserve_reverse_ip_tag=self._reserve_reverse_ip_tag,
            enable_injection=self._enable_injection)
        machine_vertex.enable_recording(self._is_recording)
        # Known issue with ReverseIPTagMulticastSourceMachineVertex
        if resources_required:
            assert (resources_required == machine_vertex.resources_required)
        return machine_vertex

    def __repr__(self):
        return self._label



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
from spinn_utilities.log import FormatAdapter
from spinn_utilities.progress_bar import ProgressBar
from spinn_front_end_common.interface.buffer_management.buffer_models \
    import (
        AbstractReceiveBuffersToHost)

logger = FormatAdapter(logging.getLogger(__name__))


class BufferExtractor(object):
    """ Extracts data in between runs.
    """

    __slots__ = []

    def __call__(self, machine_graph, placements, buffer_manager):
        """
        :param ~pacman.model.graphs.machine.MachineGraph machine_graph:
        :param ~pacman.model.placements.Placements placements:
        :param BufferManager buffer_manager:
        """
        # Count the regions to be read
        n_regions_to_read, recording_placements = self._count_regions(
            machine_graph, placements)
        if not n_regions_to_read:
            logger.info("No recorded data to extract")
            return

        # Read back the regions
        progress = ProgressBar(
            n_regions_to_read, "Extracting buffers from the last run")
        try:
            buffer_manager.get_data_for_placements(
                recording_placements, progress)
        finally:
            progress.end()

    @staticmethod
    def _count_regions(machine_graph, placements):
        """
        :param ~.MachineGraph machine_graph:
        :param ~.Placements placements:
        :rtype: tuple(int, list(~.Placement))
        """
        # Count the regions to be read
        n_regions_to_read = 0
        recording_placements = list()
        for vertex in machine_graph.vertices:
            if isinstance(vertex, AbstractReceiveBuffersToHost):
                n_regions_to_read += len(vertex.get_recorded_region_ids())
                placement = placements.get_placement_of_vertex(vertex)
                recording_placements.append(placement)
        return n_regions_to_read, recording_placements



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from spinn_utilities.progress_bar import ProgressBar
from spinn_machine import CoreSubsets
from spinn_front_end_common.abstract_models import AbstractHasAssociatedBinary
from spinn_front_end_common.utilities.utility_objs import ExecutableType


class LocateExecutableStartType(object):
    """ Discovers where applications of particular types need to be launched.
    """

    def __call__(self, graph, placements):
        """
        :param ~pacman.model.graphs.machine.MachineGraph graph:
        :param ~pacman.model.placements.Placements placements:
        :rtype: dict(ExecutableType,~spinn_machine.CoreSubsets or None)
        """
        if not graph.vertices:
            return {ExecutableType.NO_APPLICATION: None}

        binary_start_types = dict()

        progress = ProgressBar(
            graph.n_vertices, "Finding executable start types")
        for vertex in progress.over(graph.vertices):
            # try to locate binary type, but possible it doesn't have one
            if isinstance(vertex, AbstractHasAssociatedBinary):
                bin_type = vertex.get_binary_start_type()
                # update core subset with location of the vertex on the machine
                if bin_type not in binary_start_types:
                    binary_start_types[bin_type] = CoreSubsets()

                self._add_vertex_to_subset(
                    vertex, placements, binary_start_types[bin_type])

        # only got apps with no binary, such as external devices.
        # return no app
        if not binary_start_types:
            return {ExecutableType.NO_APPLICATION: None}

        return binary_start_types

    @staticmethod
    def _add_vertex_to_subset(machine_vertex, placements, core_subsets):
        """
        :param ~.MachineVertex machine_vertex:
        :param ~.Placements placements:
        :param ~.CoreSubsets core_subsets:
        """
        placement = placements.get_placement_of_vertex(machine_vertex)
        core_subsets.add_processor(
            x=placement.x, y=placement.y, processor_id=placement.p)



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from spinn_utilities.abstract_base import AbstractBase, abstractproperty
from spinn_utilities.require_subclass import require_subclass
from pacman.model.graphs.machine import MachineVertex


@require_subclass(MachineVertex)
class AbstractSupportsDatabaseInjection(object, metaclass=AbstractBase):
    """ Marks a machine vertex as supporting injection of information via a\
        database running on the controlling host.
    """

    __slots__ = ()

    @abstractproperty
    def is_in_injection_mode(self):
        """ Whether this vertex is actually in injection mode.

        :rtype: bool
        """



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os
from spinn_utilities.log import FormatAdapter

logger = FormatAdapter(logging.getLogger(__name__))

_FOLDER_NAME = "memory_map_from_processor_to_address_space"


class MemoryMapOnHostReport(object):
    """ Report on memory usage.
    """

    def __call__(
            self, report_default_directory,
            processor_to_app_data_base_address):
        """
        :param str report_default_directory:
        :param processor_to_app_data_base_address:
        :type processor_to_app_data_base_address:
            dict(tuple(int,int,int),DataWritten)
        """

        file_name = os.path.join(report_default_directory, _FOLDER_NAME)
        try:
            with open(file_name, "w") as f:
                self._describe_mem_map(f, processor_to_app_data_base_address)
        except IOError:
            logger.exception("Generate_placement_reports: Can't open file"
                             " {} for writing.", file_name)

    def _describe_mem_map(self, f, memory_map):
        f.write("On host data specification executor\n")
        for key, data in memory_map.items():
            self._describe_map_entry(f, key, data)

    @staticmethod
    def _describe_map_entry(f, key, data):
        """
        :param f:
        :param tuple(int,int,int) key:
        :param DataWritten data:
        """
        f.write(
            "{}: ('start_address': {}, hex:{}), "
            "'memory_used': {}, 'memory_written': {} \n".format(
                key, data.start_address, hex(data.start_address),
                data.memory_used, data.memory_written))



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from enum import Enum

from spinn_front_end_common.utilities.exceptions import ConfigurationException
from spinn_utilities.overrides import overrides
from pacman.executor.injection_decorator import inject_items
from pacman.model.constraints.key_allocator_constraints import (
    FixedKeyAndMaskConstraint)
from pacman.model.graphs.machine import MachineVertex, MachineEdge
from pacman.model.resources import ConstantSDRAM, ResourceContainer
from pacman.model.routing_info import BaseKeyAndMask
from spinn_front_end_common.abstract_models import (
    AbstractHasAssociatedBinary, AbstractProvidesOutgoingPartitionConstraints,
    AbstractGeneratesDataSpecification)
from spinn_front_end_common.interface.provenance import (
    ProvidesProvenanceDataFromMachineImpl)
from spinn_front_end_common.interface.simulation.simulation_utilities import (
    get_simulation_header_array)
from spinn_front_end_common.utilities.constants import (
    SYSTEM_BYTES_REQUIREMENT, SIMULATION_N_BYTES, BYTES_PER_WORD)
from spinn_front_end_common.utilities.utility_objs import ExecutableType


class CommandSenderMachineVertex(
        MachineVertex, ProvidesProvenanceDataFromMachineImpl,
        AbstractHasAssociatedBinary, AbstractGeneratesDataSpecification,
        AbstractProvidesOutgoingPartitionConstraints):
    """ Machine vertex for injecting packets at particular times or in \
        response to particular events into a SpiNNaker application.
    """
    # Regions for populations
    class DATA_REGIONS(Enum):
        SYSTEM_REGION = 0
        COMMANDS_WITH_ARBITRARY_TIMES = 1
        COMMANDS_AT_START_RESUME = 2
        COMMANDS_AT_STOP_PAUSE = 3
        PROVENANCE_REGION = 4

    # 4 for key, 4 for has payload, 4 for payload 4 for repeats, 4 for delays
    _COMMAND_WITH_PAYLOAD_SIZE = 5 * BYTES_PER_WORD

    # 4 for the time stamp
    _COMMAND_TIMESTAMP_SIZE = BYTES_PER_WORD

    # 4 for the int to represent the number of commands
    _N_COMMANDS_SIZE = BYTES_PER_WORD

    # bool for if the command has a payload (true = 1)
    _HAS_PAYLOAD = 1

    # bool for if the command does not have a payload (false = 0)
    _HAS_NO_PAYLOAD = 0

    # The name of the binary file
    BINARY_FILE_NAME = 'command_sender_multicast_source.aplx'

    # all commands will use this mask
    _DEFAULT_COMMAND_MASK = 0xFFFFFFFF

    _NOT_GOT_KEY_ERROR_MESSAGE = (
        "The command sender {} has requested key {} for outgoing "
        "partition {}, but the keys allocated to it do not match. this will "
        "cause errors in the external devices support and therefore needs "
        "fixing")

    def __init__(self, label, constraints, app_vertex=None):
        """
        :param str label: The label of this vertex
        :param constraints: Any initial constraints to this vertex
        :type constraints:
            iterable(~pacman.model.constraints.AbstractConstraint)
        :param CommandSender app_vertex:
        """
        super().__init__(label, constraints, app_vertex)

        self._timed_commands = list()
        self._commands_at_start_resume = list()
        self._commands_at_pause_stop = list()
        self._partition_id_to_keys = dict()
        self._keys_to_partition_id = dict()
        self._edge_partition_id_counter = 0
        self._vertex_to_key_map = dict()

    def add_commands(
            self, start_resume_commands, pause_stop_commands,
            timed_commands, vertex_to_send_to):
        """ Add commands to be sent down a given edge

        :param iterable(MultiCastCommand) start_resume_commands:
            The commands to send when the simulation starts or resumes from
            pause
        :param iterable(MultiCastCommand) pause_stop_commands:
            the commands to send when the simulation stops or pauses after
            running
        :param iterable(MultiCastCommand) timed_commands:
            The commands to send at specific times
        :param ~pacman.model.graphs.AbstractVertex vertex_to_send_to:
            The vertex these commands are to be sent to
        """
        # container for keys for partition mapping (remove duplicates)
        command_keys = set()
        self._vertex_to_key_map[vertex_to_send_to] = set()

        # update holders
        self._commands_at_start_resume.extend(start_resume_commands)
        self._commands_at_pause_stop.extend(pause_stop_commands)
        self._timed_commands.extend(timed_commands)

        for commands in (
                start_resume_commands, pause_stop_commands, timed_commands):
            for command in commands:
                # track keys
                command_keys.add(command.key)
                self._vertex_to_key_map[vertex_to_send_to].add(command.key)

        # create mapping between keys and partitions via partition constraint
        for key in command_keys:

            partition_id = "COMMANDS{}".format(self._edge_partition_id_counter)
            self._keys_to_partition_id[key] = partition_id
            self._partition_id_to_keys[partition_id] = key
            self._edge_partition_id_counter += 1

    @property
    @overrides(ProvidesProvenanceDataFromMachineImpl._provenance_region_id)
    def _provenance_region_id(self):
        return self.DATA_REGIONS.PROVENANCE_REGION.value

    @property
    @overrides(ProvidesProvenanceDataFromMachineImpl._n_additional_data_items)
    def _n_additional_data_items(self):
        return 0

    @property
    @overrides(MachineVertex.resources_required)
    def resources_required(self):
        sdram = (
            self.get_timed_commands_bytes() +
            self.get_n_command_bytes(self._commands_at_start_resume) +
            self.get_n_command_bytes(self._commands_at_pause_stop) +
            SYSTEM_BYTES_REQUIREMENT +
            self.get_provenance_data_size(0))

        # Return the SDRAM and 1 core
        return ResourceContainer(sdram=ConstantSDRAM(sdram))

    @inject_items({
        "machine_time_step": "MachineTimeStep",
        "time_scale_factor": "TimeScaleFactor",
        "routing_infos": "MemoryRoutingInfos"})
    @overrides(
        AbstractGeneratesDataSpecification.generate_data_specification,
        additional_arguments={
            "machine_time_step", "time_scale_factor", "routing_infos"})
    def generate_data_specification(
            self, spec, placement, machine_time_step, time_scale_factor,
            routing_infos):
        """
        :param int machine_time_step:
        :param int time_scale_factor:
        :param ~pacman.model.routing_info.RoutingInfo routing_infos:
            the routing infos
        """
        # pylint: disable=arguments-differ
        for mc_key in self._keys_to_partition_id.keys():
            allocated_mc_key = routing_infos.get_first_key_from_pre_vertex(
                self, self._keys_to_partition_id[mc_key])
            if allocated_mc_key != mc_key:
                raise ConfigurationException(
                    self._NOT_GOT_KEY_ERROR_MESSAGE.format(
                        self._label, mc_key,
                        self._keys_to_partition_id[mc_key]))

        # pylint: disable=too-many-arguments, arguments-differ
        timed_commands_size = self.get_timed_commands_bytes()
        start_resume_commands_size = \
            self.get_n_command_bytes(self._commands_at_start_resume)
        pause_stop_commands_size = \
            self.get_n_command_bytes(self._commands_at_pause_stop)

        # reverse memory regions
        self._reserve_memory_regions(
            spec, timed_commands_size, start_resume_commands_size,
            pause_stop_commands_size)

        # Write system region
        spec.comment("\n*** Spec for multicast source ***\n\n")
        spec.switch_write_focus(
            self.DATA_REGIONS.SYSTEM_REGION.value)
        spec.write_array(get_simulation_header_array(
            self.get_binary_file_name(), machine_time_step,
            time_scale_factor))

        # write commands
        spec.switch_write_focus(
            region=self.DATA_REGIONS.COMMANDS_WITH_ARBITRARY_TIMES.value)

        # write commands to spec for timed commands
        self._write_timed_commands(self._timed_commands, spec)

        # write commands fired off during a start or resume
        spec.switch_write_focus(
            region=self.DATA_REGIONS.COMMANDS_AT_START_RESUME.value)

        self._write_basic_commands(self._commands_at_start_resume, spec)

        # write commands fired off during a pause or end
        spec.switch_write_focus(
            region=self.DATA_REGIONS.COMMANDS_AT_STOP_PAUSE.value)

        self._write_basic_commands(self._commands_at_pause_stop, spec)

        # End-of-Spec:
        spec.end_specification()

    def _write_basic_commands(self, commands, spec):
        """
        :param list(MultiCastCommand) commands:
        :param ~data_specification.DataSpecificationGenerator spec:
        """
        # number of commands
        spec.write_value(len(commands))

        # write commands to region
        for command in commands:
            self._write_command(command, spec)

    def _write_timed_commands(self, timed_commands, spec):
        """
        :param list(MultiCastCommand) timed_commands:
        :param ~data_specification.DataSpecificationGenerator spec:
        """
        spec.write_value(len(timed_commands))

        # write commands
        for command in timed_commands:
            spec.write_value(command.time)
            self._write_command(command, spec)

    @classmethod
    def _write_command(cls, command, spec):
        """
        :param MultiCastCommand command:
        :param ~data_specification.DataSpecificationGenerator spec:
        """
        spec.write_value(command.key)
        if command.is_payload:
            spec.write_value(cls._HAS_PAYLOAD)
        else:
            spec.write_value(cls._HAS_NO_PAYLOAD)
        spec.write_value(command.payload if command.is_payload else 0)
        spec.write_value(command.repeat)
        spec.write_value(command.delay_between_repeats)

    def _reserve_memory_regions(
            self, spec, time_command_size, start_command_size,
            end_command_size):
        """ Reserve SDRAM space for memory areas:

        1. Area for general system information
        2. Area for timed commands
        2. Area for start commands
        3. Area for end commands

        :param ~data_specification.DataSpecificationGenerator spec:
        :param int time_command_size:
        :param int start_command_size:
        :param int end_command_size:
        :param ProvidesProvenanceDataFromMachineImpl vertex:
        """
        spec.comment("\nReserving memory space for data regions:\n\n")

        # Reserve memory:
        spec.reserve_memory_region(
            region=self.DATA_REGIONS.SYSTEM_REGION.value,
            size=SIMULATION_N_BYTES, label='system')

        spec.reserve_memory_region(
            region=self.DATA_REGIONS.COMMANDS_WITH_ARBITRARY_TIMES.value,
            size=time_command_size, label='commands with arbitrary times')

        spec.reserve_memory_region(
            region=self.DATA_REGIONS.COMMANDS_AT_START_RESUME.value,
            size=start_command_size, label='commands with start resume times')

        spec.reserve_memory_region(
            region=self.DATA_REGIONS.COMMANDS_AT_STOP_PAUSE.value,
            size=end_command_size, label='commands with stop pause times')

        self.reserve_provenance_data_region(spec)

    def get_timed_commands_bytes(self):
        """
        :rtype: int
        """
        n_bytes = self._N_COMMANDS_SIZE
        n_bytes += (
            (self._COMMAND_TIMESTAMP_SIZE + self._COMMAND_WITH_PAYLOAD_SIZE) *
            len(self._timed_commands))
        return n_bytes

    @classmethod
    def get_n_command_bytes(cls, commands):
        """
        :param list(MultiCastCommand) commands:
        :rtype: int
        """
        n_bytes = cls._N_COMMANDS_SIZE
        n_bytes += cls._COMMAND_WITH_PAYLOAD_SIZE * len(commands)
        return n_bytes

    @overrides(AbstractHasAssociatedBinary.get_binary_file_name)
    def get_binary_file_name(self):
        return self.BINARY_FILE_NAME

    @overrides(AbstractHasAssociatedBinary.get_binary_start_type)
    def get_binary_start_type(self):
        return ExecutableType.USES_SIMULATION_INTERFACE

    @overrides(AbstractProvidesOutgoingPartitionConstraints.
               get_outgoing_partition_constraints)
    def get_outgoing_partition_constraints(self, partition):
        return [FixedKeyAndMaskConstraint([
            BaseKeyAndMask(
                self._partition_id_to_keys[partition.identifier],
                self._DEFAULT_COMMAND_MASK)])]

    def _get_edges_and_partitions(self, pre_vertex, vertex_type, edge_type):
        """ Construct edges from this vertex to the vertices that this vertex\
            knows how to target (and has keys allocated for).

        .. note::
            Do not call this directly from outside either a
            :py:class:`CommandSender` or a
            :py:class:`CommandSenderMachineVertex`.

        :param pre_vertex:
        :type pre_vertex: CommandSender or CommandSenderMachineVertex
        :param type vertex_type: subclass of :py:class:`~.AbstractVertex`
        :param callable edge_type: subclass of :py:class:`~.AbstractEdge`
        :return: edges, partition IDs
        :rtype: tuple(list(~.AbstractEdge), list(str))
        """
        edges = list()
        partition_ids = list()
        keys_added = set()
        for vertex in self._vertex_to_key_map:
            if not isinstance(vertex, vertex_type):
                continue
            for key in self._vertex_to_key_map[vertex]:
                if key not in keys_added:
                    keys_added.add(key)
                    edges.append(edge_type(pre_vertex, vertex))
                    partition_ids.append(self._keys_to_partition_id[key])
        return edges, partition_ids

    def edges_and_partitions(self):
        """ Construct machine edges from this vertex to the machine vertices\
            that this vertex knows how to target (and has keys allocated for).

        :return: edges, partition IDs
        :rtype:
            tuple(list(~pacman.model.graphs.machine.MachineEdge), list(str))
        """
        return self._get_edges_and_partitions(self, MachineVertex, MachineEdge)



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from spinn_utilities.abstract_base import AbstractBase, abstractmethod


class AbstractProvidesKeyToAtomMapping(object, metaclass=AbstractBase):
    """ Interface to provide a mapping between routing key partitions and\
        atom IDs
    """

    __slots__ = ()

    @abstractmethod
    def routing_key_partition_atom_mapping(self, routing_info, partition):
        """ Returns a list of atom to key mapping.

        :param ~pacman.model.routing_info.RoutingInfo routing_info:
            the routing info object to consider
        :param ~pacman.model.graphs.OutgoingEdgePartition partition:
            the routing partition to handle.
        :return: a iterable of tuples of atom IDs to keys.
        :rtype: iterable(tuple(int,int))
        """



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import traceback
from spinn_utilities.log import FormatAdapter
from spinn_utilities.progress_bar import ProgressBar
from spinn_front_end_common.interface.provenance import (
    AbstractProvidesProvenanceDataFromMachine)

logger = FormatAdapter(logging.getLogger(__name__))


class PlacementsProvenanceGatherer(object):
    """ Gets provenance information from vertices on the machine.
    """
    __slots__ = []

    def __call__(self, transceiver, placements):
        """
        :param ~spinnman.transceiver.Transceiver transceiver:
            the SpiNNMan interface object
        :param ~pacman.model.placements.Placements placements:
            The placements of the vertices
        :rtype: list(ProvenanceDataItem)
        """
        prov_items = list()

        progress = ProgressBar(
            placements.n_placements, "Getting provenance data")

        # retrieve provenance data from any cores that provide data
        errors = list()
        for placement in progress.over(placements.placements):
            if isinstance(placement.vertex,
                          AbstractProvidesProvenanceDataFromMachine):
                # get data
                try:
                    prov_items.extend(
                        placement.vertex.get_provenance_data_from_machine(
                            transceiver, placement))
                except Exception:  # pylint: disable=broad-except
                    errors.append(traceback.format_exc())
        if errors:
            logger.warning("Errors found during provenance gathering:")
            for error in errors:
                logger.warning("{}", error)

        return prov_items



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from spinn_utilities.progress_bar import ProgressBar
from spinnman.constants import ROUTER_REGISTER_REGISTERS
from spinnman.model import DiagnosticFilter
from spinnman.model.enums import (
    DiagnosticFilterDefaultRoutingStatus, DiagnosticFilterPacketType,
    DiagnosticFilterSource)


class RoutingSetup(object):
    """ Initialises the routers. Note that this does not load any routes into\
        them.
    """
    __slots__ = []

    def __call__(self, router_tables, app_id, transceiver, machine):
        """
        :param router_tables:
        :type router_tables:
            ~pacman.model.routing_tables.MulticastRoutingTables
        :param int app_id:
        :param ~spinnman.transceiver.Transceiver transceiver:
        :param ~spinn_machine.Machine machine:
        """
        routing_tables = list(router_tables.routing_tables)
        progress = ProgressBar(routing_tables, "Preparing Routing Tables")

        # Clear the routing table for each router that needs to be set up
        # and set up the diagnostics
        for router_table in progress.over(routing_tables):
            if not machine.get_chip_at(router_table.x, router_table.y).virtual:
                transceiver.clear_multicast_routes(
                    router_table.x, router_table.y)
                transceiver.clear_router_diagnostic_counters(
                    router_table.x, router_table.y)

                # set the router diagnostic for user 3 to catch local default
                # routed packets. This can only occur when the source router
                # has no router entry, and therefore should be detected a bad
                # dropped packet.
                self._set_router_diagnostic_filters(
                    router_table.x, router_table.y, transceiver)

    @staticmethod
    def _set_router_diagnostic_filters(x, y, transceiver):
        """
        :param int x:
        :param int y:
        :param ~.Transceiver transceiver:
        """
        transceiver.set_router_diagnostic_filter(
            x, y, ROUTER_REGISTER_REGISTERS.USER_3.value,
            DiagnosticFilter(
                enable_interrupt_on_counter_event=False,
                match_emergency_routing_status_to_incoming_packet=False,
                destinations=[],
                sources=[DiagnosticFilterSource.LOCAL],
                payload_statuses=[],
                default_routing_statuses=[
                    DiagnosticFilterDefaultRoutingStatus.DEFAULT_ROUTED],
                emergency_routing_statuses=[],
                packet_types=[DiagnosticFilterPacketType.MULTICAST]))

        transceiver.set_router_diagnostic_filter(
            x, y, ROUTER_REGISTER_REGISTERS.USER_2.value,
            DiagnosticFilter(
                enable_interrupt_on_counter_event=False,
                match_emergency_routing_status_to_incoming_packet=False,
                destinations=[],
                sources=[DiagnosticFilterSource.NON_LOCAL],
                payload_statuses=[],
                default_routing_statuses=[
                    DiagnosticFilterDefaultRoutingStatus.DEFAULT_ROUTED],
                emergency_routing_statuses=[],
                packet_types=[DiagnosticFilterPacketType.MULTICAST]))



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from spinnman.processes import AbstractMultiConnectionProcess
from spinn_front_end_common.utilities.utility_objs.extra_monitor_scp_messages\
    import (
        SetRouterTimeoutMessage)


class SetRouterTimeoutProcess(AbstractMultiConnectionProcess):
    """ How to send messages to set router timeouts. These messages need to be\
        sent to cores running the extra monitor binary.

    .. note::
        SCAMP sets wait2 to zero by default!

    .. note::
        Timeouts are specified in a weird floating point format.
        See the SpiNNaker datasheet for details.
    """

    def set_wait1_timeout(self, mantissa, exponent, core_subsets):
        """ The wait1 timeout is the time from when a packet is received to\
            when emergency routing becomes enabled.

        :param int mantissa: Timeout mantissa (0 to 15)
        :param int exponent: Timeout exponent (0 to 15)
        :param ~spinn_machine.CoreSubsets core_subsets:
            Where the extra monitors that manage the routers are located.
        """
        for core_subset in core_subsets.core_subsets:
            for processor_id in core_subset.processor_ids:
                self._set_timeout(
                    core_subset, processor_id, mantissa, exponent, wait=1)

    def set_wait2_timeout(self, mantissa, exponent, core_subsets):
        """ The wait2 timeout is the time from when a packet has emergency\
            routing enabled for it to when it is dropped.

        :param int mantissa: Timeout mantissa (0 to 15)
        :param int exponent: Timeout exponent (0 to 15)
        :param ~spinn_machine.CoreSubsets core_subsets:
            Where the extra monitors that manage the routers are located.
        """
        for core_subset in core_subsets.core_subsets:
            for processor_id in core_subset.processor_ids:
                self._set_timeout(
                    core_subset, processor_id, mantissa, exponent, wait=2)

    def _set_timeout(self, core, processor_id, mantissa, exponent, wait):
        """ Set a timeout for a router controlled by an extra monitor on a\
            core. This is not a parallelised operation in order to aid\
            debugging when it fails.

        :param ~spinn_machine.CoreSubset core:
        :param int processor_id:
        :param int mantissa:
        :param int exponent:
        :param int wait:
        """
        self._send_request(SetRouterTimeoutMessage(
            core.x, core.y, processor_id,
            timeout_mantissa=mantissa, timeout_exponent=exponent, wait=wait))
        self._finish()
        self.check_for_error()



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


class ProvenanceDataItem(object):
    """ Container for provenance data
    """
    __slots__ = [
        "_message",
        "_names",
        "_report",
        "_value"]

    def __init__(self, names, value, report=False, message=""):
        """
        :param list(str) names:
            A list of strings representing the naming hierarchy of this item
        :param value: The value of the item
        :type value: int or float or str
        :param bool report: True if the item should be reported to the user
        :param str message:
            The message to send to the end user if report is True
        """
        self._names = names
        self._value = value
        self._report = report
        self._message = message

    @property
    def message(self):
        """ The message to report to the end user, or None if no message

        :rtype: str
        """
        return self._message

    @property
    def report(self):
        """ True if this provenance data entry needs reporting to the end user

        :rtype: bool
        """
        return self._report

    @property
    def names(self):
        """ The hierarchy of names of this bit of provenance data

        :rtype: list(str)
        """
        return self._names

    @property
    def value(self):
        """ The value of the item

        :rtype: int or float or str
        """
        return self._value

    def __repr__(self):
        return "{}:{}:{}:{}".format(
            self._names, self._value, self._report, self._message)

    def __str__(self):
        if self._report:
            return self._message
        return "{}: {}".format(self._names, self._value)



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os
from spinn_utilities.log import FormatAdapter
from spinn_front_end_common.utility_models import ChipPowerMonitorMachineVertex
from spinn_front_end_common.utilities.exceptions import ConfigurationException
from spinn_front_end_common.utilities.helpful_functions import (
    convert_time_diff_to_total_milliseconds)
from spinn_front_end_common.interface.interface_functions import (
    ComputeEnergyUsed)
from spinn_machine.machine import Machine

logger = FormatAdapter(logging.getLogger(__name__))


class EnergyReport(object):
    """ This class creates a report about the approximate total energy\
        consumed by a SpiNNaker job execution.
    """

    __slots__ = (
        "__report_dir", "__version", "__uses_spalloc", "__time_scale_factor")

    #: converter between joules to kilowatt hours
    JOULES_TO_KILOWATT_HOURS = 3600000

    # energy report file name
    _DETAILED_FILENAME = "detailed_energy_report.rpt"
    _SUMMARY_FILENAME = "summary_energy_report.rpt"

    def __init__(self, report_default_directory, version, spalloc_server,
                 remote_spinnaker_url, time_scale_factor):
        """
        :param str report_default_directory: location for reports
        :param int version: version of machine
        :param str spalloc_server: spalloc server IP
        :param str remote_spinnaker_url: remote SpiNNaker URL
        :param int time_scale_factor: the time scale factor
        """
        self.__report_dir = report_default_directory
        self.__version = version
        self.__uses_spalloc = bool(spalloc_server or remote_spinnaker_url)
        self.__time_scale_factor = time_scale_factor

    def write_energy_report(
            self, placements, machine, runtime, buffer_manager, power_used):
        """ Writes the report.

        :param ~pacman.model.placements.Placements placements: the placements
        :param ~spinn_machine.Machine machine: the machine
        :param int runtime:
        :param BufferManager buffer_manager:
        :param PowerUsed power_used:
        :rtype: None
        """
        # pylint: disable=too-many-arguments, too-many-locals
        if buffer_manager is None:
            logger.info("Skipping Energy report as no buffer_manager set")
            return

        # detailed report path
        detailed_report = os.path.join(
            self.__report_dir, self._DETAILED_FILENAME)

        # summary report path
        summary_report = os.path.join(
            self.__report_dir, self._SUMMARY_FILENAME)

        # figure runtime in milliseconds with time scale factor
        runtime_total_ms = runtime * self.__time_scale_factor

        # create detailed report
        with open(detailed_report, "w") as f:
            self._write_detailed_report(
                placements, machine, power_used, f, runtime_total_ms)

        # create summary report
        with open(summary_report, "w") as f:
            self._write_summary_report(runtime_total_ms, f, power_used)

    @classmethod
    def _write_summary_report(cls, runtime_total_ms, f, power_used):
        """ Write summary file

        :param int runtime_total_ms:
            Runtime with time scale factor taken into account
        :param ~io.TextIOBase f: file writer
        :param PowerUsed power_used:
        """
        # pylint: disable=too-many-arguments, too-many-locals

        # write summary data
        f.write("Summary energy file\n-------------------\n\n")
        f.write(
            "Energy used by chips during runtime is {} Joules {}\n".format(
                power_used.chip_energy_joules,
                cls.__report_time(runtime_total_ms / 1000)))
        f.write(
            "Energy used by FPGAs is {} Joules over the entire time the "
            "machine was booted {}\n".format(
                power_used.fpga_total_energy_joules,
                cls.__report_time(power_used.booted_time_secs)))
        f.write(
            "Energy used by FPGAs is {} Joules over the runtime period {}\n"
            .format(
                power_used.fpga_exec_energy_joules,
                cls.__report_time(runtime_total_ms / 1000)))
        f.write(
            "Energy used by outside router / cooling during the runtime "
            "period is {} Joules\n".format(power_used.baseline_joules))
        f.write(
            "Energy used by packet transmissions is {} Joules {}\n".format(
                power_used.packet_joules,
                cls.__report_time(power_used.total_time_secs)))
        f.write(
            "Energy used during the mapping process is {} Joules {}\n".format(
                power_used.mapping_joules,
                cls.__report_time(power_used.mapping_time_secs)))
        f.write(
            "Energy used by the data generation process is {} Joules {}\n"
            .format(
                power_used.data_gen_joules,
                cls.__report_time(power_used.data_gen_time_secs)))
        f.write(
            "Energy used during the loading process is {} Joules {}\n".format(
                power_used.loading_joules,
                cls.__report_time(power_used.loading_time_secs)))
        f.write(
            "Energy used during the data extraction process is {} Joules {}\n"
            .format(power_used.saving_joules, cls.__report_time(
                power_used.saving_time_secs)))
        f.write(
            "Total energy used by the simulation over {} milliseconds is:\n"
            "     {} Joules, or\n"
            "     {} estimated average Watts, or\n"
            "     {} kWh\n".format(
                power_used.total_time_secs * 1000,
                power_used.total_energy_joules,
                power_used.total_energy_joules / power_used.total_time_secs,
                power_used.total_energy_joules /
                cls.JOULES_TO_KILOWATT_HOURS))

    @staticmethod
    def __report_time(time):
        """
        :param float time:
        :rtype: str
        """
        if time < 1:
            return "(over {} milliseconds)".format(time * 1000)
        else:
            return "(over {} seconds)".format(time)

    def _write_detailed_report(
            self, placements, machine, power_used, f, runtime_total_ms):
        """ Write detailed report and calculate costs

        :param ~.Placements placements: placements
        :param ~.Machine machine: machine representation
        :param PowerUsed power_used:
        :param ~io.TextIOBase f: file writer
        :param float runtime_total_ms:
            total runtime with time scale factor taken into account
        """
        # pylint: disable=too-many-arguments, too-many-locals

        # write warning about accuracy etc
        self._write_warning(f)

        # figure out packet cost
        f.write("The packet cost is {} Joules\n".format(
            power_used.packet_joules))

        # figure FPGA cost over all booted and during runtime cost
        self._write_fpga_cost(power_used, f)

        # figure load time cost
        self._write_load_time_cost(power_used, f)

        # figure extraction time cost
        self._write_data_extraction_time_cost(power_used, f)

        # figure out active chips idle time
        active_chips = set()
        for placement in placements:
            if not isinstance(placement.vertex, ChipPowerMonitorMachineVertex):
                active_chips.add(machine.get_chip_at(placement.x, placement.y))
        for chip in active_chips:
            self._write_chips_active_cost(
                chip, placements, runtime_total_ms, power_used, f)

    def _write_warning(self, f):
        """ Writes the warning about this being only an estimate

        :param ~io.TextIOBase f: the writer
        :rtype: None
        """

        f.write(
            "This report is based off energy estimates for individual "
            "components of the SpiNNaker machine. It is not meant to be "
            "completely accurate. But does use provenance data gathered from "
            "the machine to estimate the energy usage and therefore should "
            "be in the right ballpark.\n\n\n")
        f.write(
            "The energy components we use are as follows:\n\n"
            "The energy usage for a chip when all cores are 100% active for "
            "a millisecond is {} Joules.\n"
            "The energy usage for a chip when all cores are not active for a "
            "millisecond is {} Joules.\n"
            "The energy used by the machine for firing a packet is {} "
            "Joules.\n"
            "The energy used by each active FPGA per millisecond is {} "
            "Joules.\n\n\n"
            .format(
                ComputeEnergyUsed.MILLIWATTS_PER_CHIP_ACTIVE_OVERHEAD,
                ComputeEnergyUsed.MILLIWATTS_PER_IDLE_CHIP,
                ComputeEnergyUsed.JOULES_PER_SPIKE,
                ComputeEnergyUsed.MILLIWATTS_PER_FPGA))

    def _write_fpga_cost(self, power_used, f):
        """ FPGA cost calculation

        :param PowerUsed power_used: the runtime
        :param ~io.TextIOBase f: the file writer
        """

        # if not spalloc, then could be any type of board
        if not self.__uses_spalloc:
            # if a spinn2 or spinn3 (4 chip boards) then they have no fpgas
            if int(self.__version) in (2, 3):
                f.write(
                    "A SpiNN-{} board does not contain any FPGA's, and so "
                    "its energy cost is 0 \n".format(self.__version))
                return
            elif int(self.__version) not in (4, 5):
                # no idea where we are; version unrecognised
                raise ConfigurationException(
                    "Do not know what the FPGA setup is for this version of "
                    "SpiNNaker machine.")

            # if a spinn4 or spinn5 board, need to verify if wrap-arounds
            # are there, if not then assume fpgas are turned off.
            if power_used.num_fpgas == 0:
                # no active fpgas
                f.write(
                    "The FPGA's on the SpiNN-{} board are turned off and "
                    "therefore the energy used by the FPGA is 0\n".format(
                        self.__version))
                return
            # active fpgas; fall through to shared main part report

        # print out as needed for spalloc and non-spalloc versions
        if self.__version is None:
            f.write(
                "{} FPGAs on the Spalloc-ed boards are turned on and "
                "therefore the energy used by the FPGA during the entire time "
                "the machine was booted (which was {} ms) is {}. "
                "The usage during execution was {}".format(
                    power_used.num_fpgas, power_used.total_time_secs * 1000,
                    power_used.fpga_total_energy_joules,
                    power_used.fpga_exec_energy_joules))
        else:
            f.write(
                "{} FPGA's on the SpiNN-{} board are turned on and "
                "therefore the energy used by the FPGA during the entire time "
                "the machine was booted (which was {} ms) is {}. "
                "The usage during execution was {}".format(
                    power_used.num_fpgas, self.__version,
                    power_used.total_time_secs * 1000,
                    power_used.fpga_total_energy_joules,
                    power_used.fpga_exec_energy_joules))

    @staticmethod
    def _write_chips_active_cost(
            chip, placements, runtime_total_ms, power_used, f):
        """ Figure out the chip active cost during simulation

        :param ~.Chip chip: the chip to consider
        :param ~.Placements placements: placements
        :param float runtime_total_ms:
        :param PowerUsed power_used:
        :param ~io.TextIOBase f: file writer
        :return: energy cost
        """

        f.write("\n")

        # detailed report print out
        for core in range(Machine.DEFAULT_MAX_CORES_PER_CHIP):
            if placements.is_processor_occupied(chip.x, chip.y, core):
                vertex = placements.get_vertex_on_processor(
                    chip.x, chip.y, core)
                label = " (running {})".format(vertex.label)
            else:
                label = ""
            energy = power_used.get_core_active_energy_joules(
                chip.x, chip.y, core)
            f.write(
                "processor {}:{}:{}{} used {} Joules of energy by "
                "being active during the execution of the simulation\n".format(
                    chip.x, chip.y, core, label, energy))

        # TAKE INTO ACCOUNT IDLE COST
        idle_cost = (
            runtime_total_ms * ComputeEnergyUsed.MILLIWATTS_PER_IDLE_CHIP)

        f.write(
            "The chip at {},{} used {} Joules of energy for by being idle "
            "during the execution of the simulation\n".format(
                chip.x, chip.y, idle_cost))

    @staticmethod
    def _write_load_time_cost(power_used, f):
        """ Energy usage from the loading phase

        :param PowerUsed power_used:
        :param ~io.TextIOBase f: file writer
        """

        # find time in milliseconds
        total_time_ms = 0.0
        for element in power_used._algorithm_timing_provenance:
            if element.names[1] == "loading":
                total_time_ms += convert_time_diff_to_total_milliseconds(
                    element.value)

        # handle active routers etc
        active_router_cost = (
            power_used.loading_time_secs * 1000 * power_used.num_frames *
            ComputeEnergyUsed.MILLIWATTS_PER_FRAME_ACTIVE_COST)

        # detailed report write
        f.write(
            "The amount of time used during the loading process is {} "
            "milliseconds.\nAssumed only 2 monitor cores is executing that "
            "this point. We also assume that there is a baseline active "
            "router/cooling component that is using {} Joules. "
            "Overall the energy usage is {} Joules.\n".format(
                total_time_ms, active_router_cost,
                power_used.loading_joules))

    @staticmethod
    def _write_data_extraction_time_cost(power_used, f):
        """ Data extraction cost

        :param PowerUsed power_used:
        :param ~io.TextIOBase f: file writer
        """

        # find time
        total_time_ms = 0.0
        for element in power_used._algorithm_timing_provenance:
            if (element.names[1] == "Execution" and element.names[2] !=
                    "run_time_of_FrontEndCommonApplicationRunner"):
                total_time_ms += convert_time_diff_to_total_milliseconds(
                    element.value)

        # handle active routers etc
        energy_cost_of_active_router = (
            total_time_ms * power_used.num_frames *
            ComputeEnergyUsed.MILLIWATTS_PER_FRAME_ACTIVE_COST)

        # detailed report
        f.write(
            "The amount of time used during the data extraction process is {} "
            "milliseconds.\nAssumed only 2 monitor cores is executing at "
            "this point. We also assume that there is a baseline active "
            "router/cooling component that is using {} Joules. Hence the "
            "overall energy usage is {} Joules.\n".format(
                total_time_ms, energy_cost_of_active_router,
                power_used.saving_joules))



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from spinn_front_end_common.utilities.helpful_functions import (
    locate_memory_region_for_placement)
from spinn_front_end_common.utilities.constants import BYTES_PER_WORD
from .profile_data import ProfileData

_PROFILE_HEADER_SIZE_BYTES = BYTES_PER_WORD
_SIZE_OF_PROFILE_DATA_ENTRY_IN_BYTES = 2 * BYTES_PER_WORD
_BYTE_OFFSET_OF_PROFILE_DATA_IN_PROFILE_REGION = BYTES_PER_WORD


def get_profile_region_size(n_samples):
    """ Get the size of the region of the profile data.

    :param int n_samples: number of different samples to record
    :return: the size in bytes used by the profile region
    :rtype: int
    """
    return _PROFILE_HEADER_SIZE_BYTES + (
        n_samples * _SIZE_OF_PROFILE_DATA_ENTRY_IN_BYTES)


def reserve_profile_region(spec, region, n_samples):
    """ Reserves the profile region for recording the profile data.

    :param ~data_specification.DataSpecificationGenerator spec:
        the DSG specification writer
    :param int region: region ID for the profile data
    :param int n_samples: number of elements being sampled
    :rtype: None
    """
    size = get_profile_region_size(n_samples)
    spec.reserve_memory_region(
        region=region, size=size, label="profilerRegion")


def write_profile_region_data(spec, region, n_samples):
    """ Writes the profile region data.

    :param ~data_specification.DataSpecificationGenerator spec:
        the DSG specification writer
    :param int region: region ID for the profile data
    :param int n_samples: number of elements being sampled
    :rtype: None
    """
    spec.switch_write_focus(region)
    spec.write_value(n_samples)


def get_profiling_data(profile_region, tag_labels, txrx, placement):
    """ Utility function to get profile data from a profile region.

    :param int profile_region: DSG region to get profiling data out of SDRAM
    :param list(str) tag_labels: labels for the profiling data
    :param ~spinnman.transceiver.Transceiver txrx: SpiNNMan transceiver
    :param ~pacman.model.placements.Placement placement: placement
    :rtype: ProfileData
    """
    profile_data = ProfileData(tag_labels)

    address = locate_memory_region_for_placement(
        placement=placement, region=profile_region, transceiver=txrx)

    # Read the profiling data size
    words_written = txrx.read_word(placement.x, placement.y, address)

    # Read the profiling data
    if words_written != 0:
        address += _BYTE_OFFSET_OF_PROFILE_DATA_IN_PROFILE_REGION
        profile_data.add_data(txrx.read_memory(
            placement.x, placement.y, address, words_written * BYTES_PER_WORD))

    return profile_data



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from collections import deque
from threading import Lock
import logging
from spinn_utilities.log import FormatAdapter
from spinnman.messages.eieio.command_messages import (
    EventStopRequest, HostSendSequencedData)
from spinn_front_end_common.utilities.exceptions import SpinnFrontEndException

logger = FormatAdapter(logging.getLogger(__name__))

#: The total number of sequence numbers
_N_SEQUENCES = 256


class BuffersSentDeque(object):
    """ A tracker of buffers sent / to send for a region
    """

    __slots__ = [
        #: The region being managed
        "_region",

        #: A queue of messages sent, ordered by sequence number
        "_buffers_sent",

        #: The current sequence number of the region
        "_sequence_number",

        #: A lock for the sequence number
        "_sequence_lock",

        #: The last sequence number to be received on the machine
        "_last_received_sequence_number",

        #: True if the stop message has been sent
        "_sent_stop_message",

        #: The number of sequence numbers allowed in a single transmission
        "_n_sequences_per_transmission"
    ]

    def __init__(self, region, sent_stop_message=False,
                 n_sequences_per_tranmission=64):
        """
        :param int region: The region being managed
        :param bool sent_stop_message: True if the stop message has been sent
        :param int n_sequences_per_tranmission:
            The number of sequences allowed in each transmission set
        """
        self._region = region
        self._buffers_sent = deque(maxlen=n_sequences_per_tranmission)
        self._sequence_number = 0
        self._sequence_lock = Lock()
        self._last_received_sequence_number = _N_SEQUENCES - 1
        self._sent_stop_message = sent_stop_message
        self._n_sequences_per_transmission = n_sequences_per_tranmission

    @property
    def is_full(self):
        """ Determine if the number of messages sent is at the limit for the\
            sequencing system.

        :rtype: bool
        """
        return len(self._buffers_sent) >= self._n_sequences_per_transmission

    def is_empty(self):
        """ Determine if there are no messages.

        :rtype: int
        """
        return len(self._buffers_sent) == 0

    def send_stop_message(self):
        """ Send a message to indicate the end of all the messages.
        """
        if not self._sent_stop_message:
            self._sent_stop_message = True
            self.add_message_to_send(EventStopRequest())

    def add_message_to_send(self, message):
        """ Add a message to send.  The message is converted to a sequenced\
            message.

        :param message: The message to be added
        :type message:
            ~spinnman.messages.eieio.abstract_messages.AbstractEIEIOMessage
        """

        # If full, raise an exception
        if self.is_full:
            raise SpinnFrontEndException("The buffer is full")

        # Create a sequenced message and update the sequence number
        self._sequence_lock.acquire()
        sequenced_message = HostSendSequencedData(
            self._region, self._sequence_number, message)
        self._sequence_number = (self._sequence_number + 1) % _N_SEQUENCES
        self._sequence_lock.release()

        # Add the sequenced message to the buffers
        self._buffers_sent.append(sequenced_message)

    @property
    def messages(self):
        """ The messages that have been added to the set.

        :rtype:
            iterable(~spinnman.messages.eieio.command_messages.HostSendSequencedData)
        """
        return self._buffers_sent

    def update_last_received_sequence_number(self, last_received_sequence_no):
        """ Updates the last received sequence number.  If the sequence number\
            is within the valid window, packets before the sequence number\
            within the window are removed, and the last received sequence\
            number is updated, thus moving the window for the next call.  If\
            the sequence number is not within the valid window, it is assumed\
            to be invalid and so is ignored.

        :param int last_received_sequence_no: The new sequence number
        :return: True if update went ahead, False if it was ignored
        :rtype: bool
        """

        # The sequence number window is between the last received and
        # the last received + window size, taking account that the end
        # of the window might wrap
        min_seq_no_acceptable = self._last_received_sequence_number
        max_seq_no_acceptable = (
            (min_seq_no_acceptable + self._n_sequences_per_transmission)
            % _N_SEQUENCES)

        if (min_seq_no_acceptable <= last_received_sequence_no <=
                max_seq_no_acceptable):

            # The sequence hasn't wrapped and the sequence is valid
            self._last_received_sequence_number = last_received_sequence_no
            self._remove_messages()
            return True
        elif max_seq_no_acceptable < min_seq_no_acceptable:

            # The sequence has wrapped
            if (0 <= last_received_sequence_no <= max_seq_no_acceptable or
                    min_seq_no_acceptable <= last_received_sequence_no <=
                    _N_SEQUENCES):

                # The sequence is in the valid range
                self._last_received_sequence_number = last_received_sequence_no
                self._remove_messages()
                return True

        # If none of the above match, the sequence is out of the window
        return False

    def _remove_messages(self):
        """ Remove messages that are no longer relevant, based on the last\
            sequence number received.
        """
        min_sequence = (self._last_received_sequence_number -
                        self._n_sequences_per_transmission)
        logger.debug("Removing buffers between {} and {}",
                     min_sequence, self._last_received_sequence_number)

        # If we are at the start of the sequence numbers, keep going back up to
        # the allowed window
        if min_sequence < 0:
            back_min_sequence = min_sequence + _N_SEQUENCES
            while (self._buffers_sent and
                    self._buffers_sent[0].sequence_no > back_min_sequence):
                logger.debug("Removing buffer with sequence {}",
                             self._buffers_sent[0].sequence_no)
                self._buffers_sent.popleft()

        # Go back through the queue until we reach the last received sequence
        while (self._buffers_sent and
                min_sequence < self._buffers_sent[0].sequence_no <=
                self._last_received_sequence_number):
            logger.debug("Removing buffer with sequence {}",
                         self._buffers_sent[0].sequence_no)
            self._buffers_sent.popleft()



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from spinn_front_end_common.utilities.exceptions import ConfigurationException


class MultiCastCommand(object):
    """ A command to be sent to a vertex.
    """

    def __init__(
            self, key, payload=None, time=None, repeat=0,
            delay_between_repeats=0):
        """
        :param int key: The key of the command
        :param payload: The payload of the command
        :type payload: int or None
        :param time: The time within the simulation at which to send the
            command, or ``None`` if this is not a timed command
        :type time: int or None
        :param int repeat:
            The number of times that the command should be repeated after
            sending it once. This could be used to ensure that the command is
            sent despite lost packets. Must be between 0 and 65535
        :param int delay_between_repeats:
            The amount of time in microseconds to wait between sending repeats
            of the same command. Must be between 0 and 65535, and must be 0 if
            repeat is 0
        :raise ConfigurationException: If the repeat or delay are out of range
        """
        # pylint: disable=too-many-arguments
        if repeat < 0 or repeat > 0xFFFF:
            raise ConfigurationException(
                "repeat must be between 0 and 65535")
        if delay_between_repeats < 0 or delay_between_repeats > 0xFFFF:
            raise ConfigurationException(
                "delay_between_repeats must be between 0 and 65535")
        if delay_between_repeats > 0 and repeat == 0:
            raise ConfigurationException(
                "If repeat is 0, delay_betweeen_repeats must be 0")

        self._time = time
        self._key = key

        self._payload = payload
        self._repeat = repeat
        self._delay_between_repeats = delay_between_repeats

    @property
    def time(self):
        """ The time within the simulation at which to send the\
            command, or None if this is not a timed command

        :rtype: int or None
        """
        return self._time

    @property
    def is_timed(self):
        """ Whether this command is a timed command.

        :rtype: bool
        """
        return self._time is not None

    @property
    def key(self):
        """
        :rtype: int
        """
        return self._key

    @property
    def repeat(self):
        """
        :rtype: int
        """
        return self._repeat

    @property
    def delay_between_repeats(self):
        """
        :rtype: int
        """
        return self._delay_between_repeats

    @property
    def payload(self):
        """ The payload of the command, or None if there is no payload.

        :rtype: int or None
        """
        return self._payload

    @payload.setter
    def payload(self, payload):
        self._payload = payload

    @property
    def is_payload(self):
        """ Whether this command has a payload. By default, this returns\
            True if the payload passed in to the constructor is not None, but\
            this can be overridden to indicate that a payload will be\
            generated, despite None being passed to the constructor

        :rtype: bool
        """
        return self._payload is not None

    def __repr__(self):
        return \
            "MultiCastCommand(time={}, key={}, payload={},"\
            " time_between_repeat={}, repeats={})".format(
                self._time, self._key, self._payload,
                self._delay_between_repeats, self._repeat)



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from spinn_utilities.progress_bar import ProgressBar
from pacman.model.resources import (
    ConstantSDRAM, CoreResource, PreAllocatedResourceContainer,
    SpecificChipSDRAMResource, SpecificBoardIPtagResource)
from spinn_front_end_common.utility_models import (
    LivePacketGatherMachineVertex)


class PreAllocateResourcesForLivePacketGatherers(object):
    """ Adds Live Packet Gatherer resources as required for a machine.
    """

    def __call__(
            self, live_packet_gatherer_parameters, machine,
            pre_allocated_resources):
        """
        :param live_packet_gatherer_parameters:
            the LPG parameters requested by the script
        :type live_packet_gatherer_parameters:
            dict(LivePacketGatherParameters,
            list(tuple(~pacman.model.graphs.AbstractVertex, list(str))))
        :param ~spinn_machine.Machine machine:
            the SpiNNaker machine as discovered
        :param pre_allocated_resources: other preallocated resources
        :type pre_allocated_resources:
            ~pacman.model.resources.PreAllocatedResourceContainer
        :return: preallocated resources
        :rtype: ~pacman.model.resources.PreAllocatedResourceContainer
        """

        progress = ProgressBar(
            len(machine.ethernet_connected_chips),
            "Preallocating resources for Live Recording")

        # store how much SDRAM the LPG uses per core
        sdram_requirement = LivePacketGatherMachineVertex.get_sdram_usage()

        # for every Ethernet connected chip, get the resources needed by the
        # live packet gatherers
        sdrams = list()
        cores = list()
        iptags = list()
        for chip in progress.over(machine.ethernet_connected_chips):
            self._add_chip_lpg_reqs(
                live_packet_gatherer_parameters, chip, sdram_requirement,
                sdrams, cores, iptags)

        # note what has been preallocated
        allocated = PreAllocatedResourceContainer(
            specific_sdram_usage=sdrams, core_resources=cores,
            specific_iptag_resources=iptags)
        allocated.extend(pre_allocated_resources)
        return allocated

    @staticmethod
    def _add_chip_lpg_reqs(
            lpg_parameters, chip, lpg_sdram, sdrams, cores, iptags):
        """
        :param lpg_parameters:
        :type lpg_parameters:
            dict(LivePacketGatherParameters,
            list(tuple(~.AbstractVertex, list(str))))
        :param ~.Chip chip:
        :param int lpg_sdram:
        :param list(~.SpecificChipSDRAMResource) sdrams:
        :param list(~.CoreResource) cores:
        :param list(~.SpecificBoardIPtagResource) iptags:
        """
        # pylint: disable=too-many-arguments
        sdram_reqs = 0
        core_reqs = 0

        for lpg_params in lpg_parameters:
            if (lpg_params.board_address is None or
                    lpg_params.board_address == chip.ip_address):
                sdram_reqs += lpg_sdram
                core_reqs += 1
                iptags.append(SpecificBoardIPtagResource(
                    board=chip.ip_address,
                    ip_address=lpg_params.hostname, port=lpg_params.port,
                    strip_sdp=lpg_params.strip_sdp, tag=lpg_params.tag,
                    traffic_identifier=(
                        LivePacketGatherMachineVertex.TRAFFIC_IDENTIFIER)))

        if sdram_reqs:
            sdrams.append(SpecificChipSDRAMResource(
                chip, ConstantSDRAM(sdram_reqs)))
        if core_reqs:
            cores.append(CoreResource(chip, core_reqs))



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from spinn_utilities.abstract_base import AbstractBase, abstractmethod
from spinn_utilities.require_subclass import require_subclass
from pacman.model.graphs.machine import MachineVertex


@require_subclass(MachineVertex)
class AbstractReceiveBuffersToHost(object, metaclass=AbstractBase):
    """ Indicates that this MachineVertex can receive buffers.
    """

    __slots__ = ()

    @abstractmethod
    def get_recorded_region_ids(self):
        """ Get the recording region IDs that have been recorded using buffering

        :return: The region numbers that have active recording
        :rtype: iterable(int)
        """

    @abstractmethod
    def get_recording_region_base_address(self, txrx, placement):
        """ Get the recording region base address

        :param ~spinnman.transceiver.Transceiver txrx: the SpiNNMan instance
        :param ~pacman.model.placements.Placement placement:
            the placement object of the core to find the address of
        :return: the base address of the recording region
        :rtype: int
        """



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .abstract_changable_after_run import AbstractChangableAfterRun
from .abstract_generates_data_specification import (
    AbstractGeneratesDataSpecification)
from .abstract_has_associated_binary import AbstractHasAssociatedBinary
from .abstract_machine_allocation_controller import (
    AbstractMachineAllocationController)
from .abstract_provides_incoming_partition_constraints import (
    AbstractProvidesIncomingPartitionConstraints)
from .abstract_provides_key_to_atom_mapping import (
    AbstractProvidesKeyToAtomMapping)
from .abstract_provides_outgoing_partition_constraints import (
    AbstractProvidesOutgoingPartitionConstraints)
from .abstract_rewrites_data_specification import (
    AbstractRewritesDataSpecification)
from .abstract_send_me_multicast_commands_vertex import (
    AbstractSendMeMulticastCommandsVertex)
from .abstract_vertex_with_dependent_vertices import (
    AbstractVertexWithEdgeToDependentVertices)
from .abstract_supports_database_injection import (
    AbstractSupportsDatabaseInjection)
from .abstract_supports_bit_field_generation import (
    AbstractSupportsBitFieldGeneration)
from .abstract_supports_bit_field_routing_compression import (
    AbstractSupportsBitFieldRoutingCompression)
from .abstract_can_reset import AbstractCanReset

__all__ = ["AbstractChangableAfterRun", "AbstractGeneratesDataSpecification",
           "AbstractHasAssociatedBinary",
           "AbstractMachineAllocationController",
           "AbstractProvidesIncomingPartitionConstraints",
           "AbstractProvidesKeyToAtomMapping",
           "AbstractProvidesOutgoingPartitionConstraints",
           "AbstractRewritesDataSpecification",
           "AbstractSendMeMulticastCommandsVertex",
           "AbstractSupportsDatabaseInjection",
           "AbstractVertexWithEdgeToDependentVertices", "AbstractCanReset",
           "AbstractSupportsBitFieldGeneration",
           "AbstractSupportsBitFieldRoutingCompression"]



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from spinn_utilities.progress_bar import ProgressBar
from pacman.model.routing_tables import MulticastRoutingTables
from pacman.model.routing_tables.compressed_multicast_routing_table import (
    CompressedMulticastRoutingTable)


class ReadRoutingTablesFromMachine(object):
    """ Reads compressed routing tables from a SpiNNaker machine.
    """

    def __call__(self, transceiver, routing_tables, app_id):
        """
        :param ~spinnman.transceiver.Transceiver transceiver:
        :param routing_tables: uncompressed routing tables
        :type routing_tables:
            ~pacman.model.routing_tables.MulticastRoutingTables
        :param int app_id:
        :rtype: ~pacman.model.routing_tables.MulticastRoutingTables
        """

        progress = ProgressBar(
            routing_tables, "Reading Routing Tables from Machine")
        machine_routing_tables = MulticastRoutingTables()
        for routing_table in progress.over(routing_tables):
            # get multicast entries from machine
            machine_routing_table = self._read_routing_table(
                transceiver, routing_table, app_id)
            machine_routing_tables.add_routing_table(machine_routing_table)

        return machine_routing_tables

    @staticmethod
    def _read_routing_table(transceiver, table, app_id):
        """
        :param ~spinnman.transceiver.Transceiver transceiver:
        :param ~.UnCompressedMulticastRoutingTable table:
        :param int app_id:
        """
        machine_routing_table = \
            CompressedMulticastRoutingTable(table.x, table.y)
        for routing_entry in transceiver.get_multicast_routes(
                table.x, table.y, app_id):
            machine_routing_table.add_multicast_routing_entry(routing_entry)
        return machine_routing_table



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import math
import struct
import numpy
from enum import IntEnum
from spinn_utilities.log import FormatAdapter
from spinn_utilities.overrides import overrides
from spinnman.messages.eieio import EIEIOPrefix, EIEIOType
from spinnman.messages.eieio.data_messages import EIEIODataHeader
from pacman.executor.injection_decorator import (
    inject_items, supports_injection, inject)
from pacman.model.constraints.key_allocator_constraints import (
    FixedKeyAndMaskConstraint)
from pacman.model.constraints.placer_constraints import BoardConstraint
from pacman.model.resources import (
    CPUCyclesPerTickResource, DTCMResource,
    ReverseIPtagResource, ResourceContainer, VariableSDRAM)
from pacman.model.routing_info import BaseKeyAndMask
from pacman.model.graphs.common import Slice
from pacman.model.graphs.machine import MachineVertex
from spinn_front_end_common.utilities.helpful_functions import (
    locate_memory_region_for_placement)
from spinn_front_end_common.interface.buffer_management.buffer_models import (
    SendsBuffersFromHostPreBufferedImpl, AbstractReceiveBuffersToHost)
from spinn_front_end_common.interface.buffer_management.storage_objects\
    .buffered_sending_region import (
        get_n_bytes)
from spinn_front_end_common.utilities import globals_variables
from spinn_front_end_common.interface.buffer_management.storage_objects \
    import (
        BufferedSendingRegion)
from spinn_front_end_common.utilities.constants import (
    SDP_PORTS, SYSTEM_BYTES_REQUIREMENT, SIMULATION_N_BYTES, BYTES_PER_WORD,
    MICRO_TO_MILLISECOND_CONVERSION)
from spinn_front_end_common.utilities.exceptions import ConfigurationException
from spinn_front_end_common.abstract_models import (
    AbstractProvidesOutgoingPartitionConstraints,
    AbstractGeneratesDataSpecification, AbstractHasAssociatedBinary,
    AbstractSupportsDatabaseInjection)
from spinn_front_end_common.interface.simulation.simulation_utilities import (
    get_simulation_header_array)
from spinn_front_end_common.interface.provenance import (
    ProvidesProvenanceDataFromMachineImpl)
from spinn_front_end_common.interface.buffer_management.recording_utilities \
    import (get_recording_header_array, get_recording_header_size,
            get_recording_data_constant_size)
from spinn_front_end_common.utilities.utility_objs import (
    ProvenanceDataItem, ExecutableType)

logger = FormatAdapter(logging.getLogger(__name__))

_DEFAULT_MALLOC_REGIONS = 2
_ONE_WORD = struct.Struct("<I")
_TWO_SHORTS = struct.Struct("<HH")

# The microseconds per timestep will be divided by this for the max offset
_MAX_OFFSET_DENOMINATOR = 10


@supports_injection
class ReverseIPTagMulticastSourceMachineVertex(
        MachineVertex, AbstractGeneratesDataSpecification,
        AbstractHasAssociatedBinary, AbstractSupportsDatabaseInjection,
        ProvidesProvenanceDataFromMachineImpl,
        AbstractProvidesOutgoingPartitionConstraints,
        SendsBuffersFromHostPreBufferedImpl, AbstractReceiveBuffersToHost):
    """ A model which allows events to be injected into SpiNNaker and\
        converted in to multicast packets.

    :param str label: The label of this vertex
    :param vertex_slice:
        The slice served via this multicast source
    :type vertex_slice: ~pacman.model.graphs.common.Slice or None
    :param app_vertex:
        The associated application vertex
    :type app_vertex: ReverseIpTagMultiCastSource or None
    :param int n_keys: The number of keys to be sent via this mulitcast source
        (can't be None if vertex_slice is also None)
    :param iterable(~pacman.model.constraints.AbstractConstraint) constraints:
        Any initial constraints to this vertex
    :param str board_address:
        The IP address of the board on which to place this vertex if receiving
        data, either buffered or live (by default, any board is chosen)
    :param int receive_port:
        The port on the board that will listen for incoming event packets
        (default is to disable this feature; set a value to enable it, or set
        the `reserve_reverse_ip_tag parameter` to True if a random port is to
        be used)
    :param int receive_sdp_port:
        The SDP port to listen on for incoming event packets (defaults to 1)
    :param int receive_tag:
        The IP tag to use for receiving live events (uses any by default)
    :param float receive_rate:
    :param int virtual_key:
        The base multicast key to send received events with (assigned
        automatically by default)
    :param int prefix:
        The prefix to "or" with generated multicast keys (default is no prefix)
    :param ~spinnman.messages.eieio.EIEIOPrefix prefix_type:
        Whether the prefix should apply to the upper or lower half of the
        multicast keys (default is upper half)
    :param bool check_keys:
        True if the keys of received events should be verified before sending
        (default False)
    :param ~numpy.ndarray send_buffer_times:
        An array of arrays of time steps at which keys should be sent (one
        array for each key, default disabled)
    :param str send_buffer_partition_id:
        The ID of the partition containing the edges down which the events are
        to be sent
    :param bool reserve_reverse_ip_tag:
        True if the source should set up a tag through which it can receive
        packets; if port is set to None this can be used to enable the
        reception of packets on a randomly assigned port, which can be read
        from the database
    :param bool enable_injection:
        Flag to indicate that data will be received to inject
    """

    class _REGIONS(IntEnum):
        SYSTEM = 0
        CONFIGURATION = 1
        RECORDING = 2
        SEND_BUFFER = 3
        PROVENANCE_REGION = 4

    class _PROVENANCE_ITEMS(IntEnum):
        N_RECEIVED_PACKETS = 0
        N_SENT_PACKETS = 1
        INCORRECT_KEYS = 2
        INCORRECT_PACKETS = 3
        LATE_PACKETS = 4

    # 13 ints (1. has prefix, 2. prefix, 3. prefix type, 4. check key flag,
    #          5. has key, 6. key, 7. mask, 8. buffer space,
    #          9. send buffer flag before notify, 10. tag,
    #          11. tag destination (y, x), 12. receive SDP port,
    #          13. timer offset)
    _CONFIGURATION_REGION_SIZE = 13 * BYTES_PER_WORD

    # Counts to do timer offsets
    _n_vertices = 0
    _n_data_specs = 0

    def __init__(
            self, label,
            vertex_slice=None,
            app_vertex=None,
            n_keys=None,
            constraints=None,

            # General input and output parameters
            board_address=None,

            # Live input parameters
            receive_port=None,
            receive_sdp_port=SDP_PORTS.INPUT_BUFFERING_SDP_PORT.value,
            receive_tag=None,
            receive_rate=10,

            # Key parameters
            virtual_key=None, prefix=None,
            prefix_type=None, check_keys=False,

            # Send buffer parameters
            send_buffer_times=None,
            send_buffer_partition_id=None,

            # Extra flag for receiving packets without a port
            reserve_reverse_ip_tag=False,

            # Flag to indicate that data will be received to inject
            enable_injection=False):
        # pylint: disable=too-many-arguments, too-many-locals
        if vertex_slice is None:
            if n_keys is not None:
                vertex_slice = Slice(0, n_keys - 1)
            else:
                raise KeyError("Either provide a vertex_slice or n_keys")

        super().__init__(label, constraints, app_vertex, vertex_slice)

        self._reverse_iptags = None
        self._n_keys = vertex_slice.n_atoms

        # Set up for receiving live packets
        if receive_port is not None or reserve_reverse_ip_tag:
            self._reverse_iptags = [ReverseIPtagResource(
                port=receive_port, sdp_port=receive_sdp_port,
                tag=receive_tag)]
            if board_address is not None:
                self.add_constraint(BoardConstraint(board_address))
        self._receive_rate = receive_rate
        self._receive_sdp_port = receive_sdp_port

        # Work out if buffers are being sent
        self._send_buffer = None
        self._send_buffer_partition_id = send_buffer_partition_id
        self._send_buffer_size = 0
        n_buffer_times = 0
        if send_buffer_times is not None:
            for i in send_buffer_times:
                if hasattr(i, "__len__"):
                    n_buffer_times += len(i)
                else:
                    # assuming this must be a single integer
                    n_buffer_times += 1
            if n_buffer_times == 0:
                logger.warning(
                    "Combination of send_buffer_times {} and slice {} results "
                    "in a core with a ReverseIPTagMulticastSourceMachineVertex"
                    " which does not spike", send_buffer_times, vertex_slice)
        if n_buffer_times == 0:
            self._send_buffer_times = None
            self._send_buffers = None
        else:
            self._install_send_buffer(send_buffer_times)

        # Set up for recording (if requested)
        self._is_recording = False

        # set flag for checking if in injection mode
        self._in_injection_mode = (
            receive_port is not None or reserve_reverse_ip_tag or
            enable_injection)

        # Sort out the keys to be used
        self._virtual_key = virtual_key
        self._mask = None
        self._prefix = prefix
        self._prefix_type = prefix_type
        self._check_keys = check_keys

        # Work out the prefix details
        if self._prefix is not None:
            if self._prefix_type is None:
                self._prefix_type = EIEIOPrefix.UPPER_HALF_WORD
            if self._prefix_type == EIEIOPrefix.UPPER_HALF_WORD:
                self._prefix = prefix << 16

        # If the user has specified a virtual key
        if self._virtual_key is not None:
            self._install_virtual_key(vertex_slice.n_atoms)

        self._n_vertices += 1

    @staticmethod
    def _max_send_buffer_keys_per_timestep(send_buffer_times, n_keys):
        """
        :param send_buffer_times: When events will be sent
        :type send_buffer_times:
            ~numpy.ndarray(~numpy.ndarray(numpy.int32)) or
            list(~numpy.ndarray(numpy.int32)) or None
        :param int n_keys:
        :rtype: int
        """
        if len(send_buffer_times) and hasattr(send_buffer_times[0], "__len__"):
            counts = numpy.bincount(numpy.concatenate(send_buffer_times))
            if len(counts):
                return max(counts)
            return 0
        if len(send_buffer_times):
            counts = numpy.bincount(send_buffer_times)
            if len(counts):
                return n_keys * max(counts)
            return 0
        return 0

    @classmethod
    def _send_buffer_sdram_per_timestep(cls, send_buffer_times, n_keys):
        """ Determine the amount of SDRAM required per timestep.

        :param send_buffer_times:
        :type send_buffer_times:
            ~numpy.ndarray(~numpy.ndarray(numpy.int32)) or
            list(~numpy.ndarray(numpy.int32)) or None
        :param int n_keys:
        :rtype: int
        """
        # If there is a send buffer, calculate the keys per timestep
        if send_buffer_times is not None:
            return get_n_bytes(cls._max_send_buffer_keys_per_timestep(
                send_buffer_times, n_keys))
        return 0

    @classmethod
    def _recording_sdram_per_timestep(
            cls, machine_time_step, is_recording, receive_rate,
            send_buffer_times, n_keys):
        """
        :param int machine_time_step:
        :param bool is_recording:
        :param float receive_rate:
        :param send_buffer_times:
        :type send_buffer_times:
            ~numpy.ndarray(~numpy.ndarray(numpy.int32)) or
            list(~numpy.ndarray(numpy.int32)) or None
        :param int n_keys:
        :rtype: int
        """
        # If not recording, no SDRAM needed per timestep
        if not is_recording:
            return 0

        # If recording send data, the recorded size is the send size
        if send_buffer_times is not None:
            return cls._send_buffer_sdram_per_timestep(
                send_buffer_times, n_keys)

        # Recording live data, use the user provided receive rate
        keys_per_timestep = math.ceil(
            receive_rate / (
                machine_time_step * MICRO_TO_MILLISECOND_CONVERSION) * 1.1)
        header_size = EIEIODataHeader.get_header_size(
            EIEIOType.KEY_32_BIT, is_payload_base=True)
        # Maximum size is one packet per key
        return ((header_size + EIEIOType.KEY_32_BIT.key_bytes) *
                keys_per_timestep)

    def _install_send_buffer(self, send_buffer_times):
        """
        :param ~numpy.ndarray send_buffer_times:
        """
        if len(send_buffer_times) and hasattr(send_buffer_times[0], "__len__"):
            # Working with a list of lists so check length
            if len(send_buffer_times) != self._n_keys:
                raise ConfigurationException(
                    "The array or arrays of times {} does not have the "
                    "expected length of {}".format(
                        send_buffer_times, self._n_keys))

        self._send_buffer = BufferedSendingRegion()
        self._send_buffer_times = send_buffer_times
        self._send_buffers = {
            self._REGIONS.SEND_BUFFER: self._send_buffer
        }

    def _install_virtual_key(self, n_keys):
        """
        :param int n_keys:
        """
        # check that virtual key is valid
        if self._virtual_key < 0:
            raise ConfigurationException("Virtual keys must be positive")

        # Get a mask and maximum number of keys for the number of keys
        # requested
        self._mask = self._calculate_mask(n_keys)

        if self._prefix is not None:
            # Check that the prefix doesn't change the virtual key in the
            # masked area
            masked_key = (self._virtual_key | self._prefix) & self._mask
            if self._virtual_key != masked_key:
                raise ConfigurationException(
                    "The number of keys, virtual key and key prefix settings "
                    "don't work together")
        else:
            # If no prefix was generated, generate one
            self._prefix_type = EIEIOPrefix.UPPER_HALF_WORD
            self._prefix = self._virtual_key

    @property
    @overrides(ProvidesProvenanceDataFromMachineImpl._provenance_region_id)
    def _provenance_region_id(self):
        return self._REGIONS.PROVENANCE_REGION

    @property
    @overrides(ProvidesProvenanceDataFromMachineImpl._n_additional_data_items)
    def _n_additional_data_items(self):
        return 5

    @property
    @overrides(MachineVertex.resources_required)
    def resources_required(self):
        sim = globals_variables.get_simulator()
        sdram = self.get_sdram_usage(
            self._send_buffer_times, self._is_recording,
            sim.machine_time_step, self._receive_rate, self._n_keys)

        resources = ResourceContainer(
            dtcm=DTCMResource(self.get_dtcm_usage()),
            sdram=sdram,
            cpu_cycles=CPUCyclesPerTickResource(self.get_cpu_usage()),
            reverse_iptags=self._reverse_iptags)
        return resources

    @classmethod
    def get_sdram_usage(
            cls, send_buffer_times, recording_enabled, machine_time_step,
            receive_rate, n_keys):
        """
        :param send_buffer_times: When events will be sent
        :type send_buffer_times:
            ~numpy.ndarray(~numpy.ndarray(numpy.int32)) or
            list(~numpy.ndarray(numpy.int32)) or None
        :param bool recording_enabled: Whether recording is done
        :param int machine_time_step: What the machine timestep is
        :param float receive_rate: What the expected message receive rate is
        :param int n_keys: How many keys are being sent
        :rtype: ~pacman.model.resources.VariableSDRAM
        """
        static_usage = (
            SYSTEM_BYTES_REQUIREMENT +
            cls._CONFIGURATION_REGION_SIZE +
            get_recording_header_size(1) +
            get_recording_data_constant_size(1) +
            cls.get_provenance_data_size(0))
        per_timestep = (
            cls._send_buffer_sdram_per_timestep(send_buffer_times, n_keys) +
            cls._recording_sdram_per_timestep(
                machine_time_step, recording_enabled, receive_rate,
                send_buffer_times, n_keys))
        static_usage += per_timestep
        return VariableSDRAM(static_usage, per_timestep)

    @staticmethod
    def get_dtcm_usage():
        return 1

    @staticmethod
    def get_cpu_usage():
        return 1

    @staticmethod
    def _n_regions_to_allocate(send_buffering, recording):
        """ Get the number of regions that will be allocated

        :param bool send_buffering:
        :param bool recording:
        :rtype: int
        """
        if recording and send_buffering:
            return 5
        elif recording or send_buffering:
            return 4
        return 3

    @property
    def send_buffer_times(self):
        """ When events will be sent.

        :rtype:
            ~numpy.ndarray(~numpy.ndarray(numpy.int32)) or
            list(~numpy.ndarray(numpy.int32)) or None
        """
        return self._send_buffer_times

    @send_buffer_times.setter
    def send_buffer_times(self, send_buffer_times):
        """
        :type send_buffer_times:
            ~numpy.ndarray(~numpy.ndarray(numpy.int32)) or
            list(~numpy.ndarray(numpy.int32)) or None
        """
        self._install_send_buffer(send_buffer_times)

    @staticmethod
    def _is_in_range(
            time_stamp_in_ticks,
            first_machine_time_step, n_machine_time_steps):
        """
        :param int time_stamp_in_ticks:
        :param int first_machine_time_step:
        :param n_machine_time_steps:
        :type n_machine_time_steps: int or None
        """
        return (n_machine_time_steps is None) or (
            first_machine_time_step <= time_stamp_in_ticks <
            n_machine_time_steps)

    def _fill_send_buffer(
            self, first_machine_time_step, run_until_timesteps):
        """ Fill the send buffer with keys to send.

        :param int first_machine_time_step:
        :param int run_until_timesteps:
        """
        key_to_send = self._virtual_key
        if self._virtual_key is None:
            key_to_send = 0

        if self._send_buffer is not None:
            self._send_buffer.clear()
        if (self._send_buffer_times is not None and
                len(self._send_buffer_times)):
            if hasattr(self._send_buffer_times[0], "__len__"):
                # Works with a list-of-lists
                self.__fill_send_buffer_2d(
                    key_to_send, first_machine_time_step, run_until_timesteps)
            else:
                # Work with a single list
                self.__fill_send_buffer_1d(
                    key_to_send, first_machine_time_step, run_until_timesteps)

    def __fill_send_buffer_2d(
            self, key_base, first_time_step, n_time_steps):
        """
        :param int key_base:
        :param int first_time_step:
        :param int n_time_steps:
        """
        for key in range(self._n_keys):
            for tick in sorted(self._send_buffer_times[key]):
                if self._is_in_range(tick, first_time_step, n_time_steps):
                    self._send_buffer.add_key(tick, key_base + key)

    def __fill_send_buffer_1d(
            self, key_base, first_time_step, n_time_steps):
        """
        :param int key_base:
        :param int first_time_step:
        :param int n_time_steps:
        """
        key_list = [key + key_base for key in range(self._n_keys)]
        for tick in sorted(self._send_buffer_times):
            if self._is_in_range(tick, first_time_step, n_time_steps):
                self._send_buffer.add_keys(tick, key_list)

    @staticmethod
    def _generate_prefix(virtual_key, prefix_type):
        """
        :param ~.EIEIOPrefix prefix_type:
        :param int virtual_key:
        :rtype: int
        """
        if prefix_type == EIEIOPrefix.LOWER_HALF_WORD:
            return virtual_key & 0xFFFF
        return (virtual_key >> 16) & 0xFFFF

    @staticmethod
    def _calculate_mask(n_neurons):
        """
        :param int n_neurons:
        :rtype: int
        """
        temp_value = n_neurons.bit_length()
        max_key = 2**temp_value - 1
        mask = 0xFFFFFFFF - max_key
        return mask

    def enable_recording(self, new_state=True):
        """ Enable recording of the keys sent.

        :param bool new_state:
        """
        self._is_recording = new_state

    def _reserve_regions(self, spec, n_machine_time_steps):
        """
        :param ~.DataSpecificationGenerator spec:
        :param int n_machine_time_steps:
        """
        # Reserve system and configuration memory regions:
        spec.reserve_memory_region(
            region=self._REGIONS.SYSTEM,
            size=SIMULATION_N_BYTES, label='SYSTEM')
        spec.reserve_memory_region(
            region=self._REGIONS.CONFIGURATION,
            size=self._CONFIGURATION_REGION_SIZE, label='CONFIGURATION')

        # Reserve recording buffer regions if required
        spec.reserve_memory_region(
            region=self._REGIONS.RECORDING,
            size=get_recording_header_size(1),
            label="RECORDING")

        # Reserve send buffer region if required
        if self._send_buffer_times is not None:
            self._send_buffer_size = (
                self._send_buffer_sdram_per_timestep(
                    self._send_buffer_times, self._n_keys) *
                n_machine_time_steps)
            if self._send_buffer_size:
                spec.reserve_memory_region(
                    region=self._REGIONS.SEND_BUFFER,
                    size=self._send_buffer_size, label="SEND_BUFFER",
                    empty=True)

        self.reserve_provenance_data_region(spec)

    def _update_virtual_key(self, routing_info, machine_graph):
        """
        :param ~pacman.model.routing_info.RoutingInfo routing_info:
        :param ~pacman.model.graphs.machine.MachineGraph machine_graph:
        """
        if self._virtual_key is None:
            if self._send_buffer_partition_id is not None:
                rinfo = routing_info.get_routing_info_from_pre_vertex(
                    self, self._send_buffer_partition_id)

                # if no edge leaving this vertex, no key needed
                if rinfo is not None:
                    self._virtual_key = rinfo.first_key
                    self._mask = rinfo.first_mask
            else:
                partitions = machine_graph\
                    .get_multicast_edge_partitions_starting_at_vertex(self)
                partition = next(iter(partitions), None)

                if partition is not None:
                    rinfo = routing_info.get_routing_info_from_partition(
                        partition)
                    self._virtual_key = rinfo.first_key
                    self._mask = rinfo.first_mask

        if self._virtual_key is not None and self._prefix is None:
            self._prefix_type = EIEIOPrefix.UPPER_HALF_WORD
            self._prefix = self._virtual_key

    def _write_configuration(
            self, spec, machine_time_step, time_scale_factor):
        """
        :param ~.DataSpecificationGenerator spec:
        :param int machine_time_step:
        :param int time_scale_factor:
        """
        spec.switch_write_focus(region=self._REGIONS.CONFIGURATION)

        # Write apply_prefix and prefix and prefix_type
        if self._prefix is None:
            spec.write_value(data=0)
            spec.write_value(data=0)
            spec.write_value(data=0)
        else:
            spec.write_value(data=1)
            spec.write_value(data=self._prefix)
            spec.write_value(data=self._prefix_type.value)

        # Write check
        if self._check_keys:
            spec.write_value(data=1)
        else:
            spec.write_value(data=0)

        # Write if you have a key to transmit write it and the mask,
        # otherwise write flags to fill in space
        if self._virtual_key is None:
            spec.write_value(data=0)
            spec.write_value(data=0)
            spec.write_value(data=0)
        else:
            spec.write_value(data=1)
            spec.write_value(data=self._virtual_key)
            spec.write_value(data=self._mask)

        # Write send buffer data
        if self._send_buffer_times is not None:

            spec.write_value(data=self._send_buffer_size)
            # The following disable the buffer notifications
            spec.write_value(data=self._send_buffer_size + 256)
            spec.write_value(data=0)
            spec.write_value(_ONE_WORD.unpack(_TWO_SHORTS.pack(0, 0))[0])
        else:
            spec.write_value(data=0)
            spec.write_value(data=0)
            spec.write_value(data=0)
            spec.write_value(data=0)

        # write SDP port to which SDP packets will be received
        spec.write_value(data=self._receive_sdp_port)

        # write timer offset in microseconds
        max_offset = ((
            machine_time_step * time_scale_factor) // (
            _MAX_OFFSET_DENOMINATOR * 2))
        spec.write_value(
            (int(math.ceil(max_offset / self._n_vertices)) *
             ReverseIPTagMulticastSourceMachineVertex._n_data_specs) +
            int(math.ceil(max_offset)))
        ReverseIPTagMulticastSourceMachineVertex._n_data_specs += 1

    @inject_items({
        "machine_time_step": "MachineTimeStep",
        "time_scale_factor": "TimeScaleFactor",
        "machine_graph": "MemoryMachineGraph",
        "routing_info": "MemoryRoutingInfos",
        "first_machine_time_step": "FirstMachineTimeStep",
        "data_n_time_steps": "DataNTimeSteps",
        "run_until_timesteps": "RunUntilTimeSteps"
    })
    @overrides(
        AbstractGeneratesDataSpecification.generate_data_specification,
        additional_arguments={
            "machine_time_step", "time_scale_factor", "machine_graph",
            "routing_info", "first_machine_time_step",
            "data_n_time_steps", "run_until_timesteps"
        })
    def generate_data_specification(
            self, spec, placement,  # @UnusedVariable
            machine_time_step, time_scale_factor, machine_graph, routing_info,
            first_machine_time_step, data_n_time_steps, run_until_timesteps):
        """
        :param int machine_time_step:
        :param int time_scale_factor:
        :param ~pacman.model.graphs.machine.MachineGraph machine_graph:
        :param ~pacman.model.routing_info.RoutingInfo routing_info:
        :param int first_machine_time_step:
        :param int data_n_time_steps:
        :param int run_until_timesteps:
        """
        # pylint: disable=too-many-arguments, arguments-differ
        self._update_virtual_key(routing_info, machine_graph)
        self._fill_send_buffer(first_machine_time_step, run_until_timesteps)

        # Reserve regions
        self._reserve_regions(spec, data_n_time_steps)

        # Write the system region
        spec.switch_write_focus(self._REGIONS.SYSTEM)
        spec.write_array(get_simulation_header_array(
            self.get_binary_file_name(), machine_time_step,
            time_scale_factor))

        # Write the additional recording information
        spec.switch_write_focus(self._REGIONS.RECORDING)
        recording_size = 0
        if self._is_recording:
            per_timestep = self._recording_sdram_per_timestep(
                machine_time_step, self._is_recording, self._receive_rate,
                self._send_buffer_times, self._n_keys)
            recording_size = per_timestep * data_n_time_steps
        spec.write_array(get_recording_header_array([recording_size]))

        # Write the configuration information
        self._write_configuration(
            spec, machine_time_step, time_scale_factor)

        # End spec
        spec.end_specification()

    @overrides(AbstractHasAssociatedBinary.get_binary_file_name)
    def get_binary_file_name(self):
        return "reverse_iptag_multicast_source.aplx"

    @overrides(AbstractHasAssociatedBinary.get_binary_start_type)
    def get_binary_start_type(self):
        return ExecutableType.USES_SIMULATION_INTERFACE

    @overrides(AbstractProvidesOutgoingPartitionConstraints.
               get_outgoing_partition_constraints)
    def get_outgoing_partition_constraints(self, partition):  # @UnusedVariable
        if self._virtual_key is not None:
            return list([FixedKeyAndMaskConstraint(
                [BaseKeyAndMask(self._virtual_key, self._mask)])])
        return list()

    @property
    def virtual_key(self):
        """
        :rtype: int or None
        """
        return self._virtual_key

    @property
    def mask(self):
        """
        :rtype: int or None
        """
        return self._mask

    @property
    @overrides(AbstractSupportsDatabaseInjection.is_in_injection_mode)
    def is_in_injection_mode(self):
        return self._in_injection_mode

    @inject("FirstMachineTimeStep")
    @inject_items({
        "run_until_timesteps": "RunUntilTimeSteps"
    })
    def update_buffer(
            self, first_machine_time_step, run_until_timesteps):
        """ Updates the buffers on specification of the first machine timestep.
            Note: This is called by injection.

        :param int first_machine_time_step:
            The first machine time step in the simulation
        :param int run_until_timesteps:
            The last machine time step in the simulation
        """
        if self._virtual_key is not None:
            self._fill_send_buffer(
                first_machine_time_step, run_until_timesteps)

    @overrides(AbstractReceiveBuffersToHost.get_recorded_region_ids)
    def get_recorded_region_ids(self):
        if not self._is_recording:
            return []
        return [0]

    @overrides(AbstractReceiveBuffersToHost.get_recording_region_base_address)
    def get_recording_region_base_address(self, txrx, placement):
        return locate_memory_region_for_placement(
            placement, self._REGIONS.RECORDING, txrx)

    @property
    def send_buffers(self):
        """
        :rtype: dict(int,BufferedSendingRegion)
        """
        return self._send_buffers

    @send_buffers.setter
    def send_buffers(self, value):
        self._send_buffers = value

    def get_region_buffer_size(self, region):
        """
        :param int region: Region ID
        :return: Size of buffer, in bytes.
        :rtype: int
        """
        if region == self._REGIONS.SEND_BUFFER:
            return self._send_buffer_size
        return 0

    @overrides(ProvidesProvenanceDataFromMachineImpl.
               get_provenance_data_from_machine)
    def get_provenance_data_from_machine(self, transceiver, placement):
        provenance_data = self._read_provenance_data(transceiver, placement)
        provenance_items = self._read_basic_provenance_items(
            provenance_data, placement)
        provenance_data = self._get_remaining_provenance_data_items(
            provenance_data)
        _, _, _, _, names = self._get_placement_details(placement)

        provenance_items.append(ProvenanceDataItem(
            self._add_name(names, "received_sdp_packets"),
            provenance_data[self._PROVENANCE_ITEMS.N_RECEIVED_PACKETS],
            report=(
                provenance_data[
                    self._PROVENANCE_ITEMS.N_RECEIVED_PACKETS] == 0 and
                self._send_buffer_times is None),
            message=(
                "No SDP packets were received by {}.  If you expected packets"
                " to be injected, this could indicate an error".format(
                    self._label))))
        provenance_items.append(ProvenanceDataItem(
            self._add_name(names, "send_multicast_packets"),
            provenance_data[self._PROVENANCE_ITEMS.N_SENT_PACKETS],
            report=provenance_data[self._PROVENANCE_ITEMS.N_SENT_PACKETS] == 0,
            message=(
                "No multicast packets were sent by {}.  If you expected"
                " packets to be sent this could indicate an error".format(
                    self._label))))
        provenance_items.append(ProvenanceDataItem(
            self._add_name(names, "incorrect_keys"),
            provenance_data[self._PROVENANCE_ITEMS.INCORRECT_KEYS],
            report=provenance_data[self._PROVENANCE_ITEMS.INCORRECT_KEYS] > 0,
            message=(
                "Keys were received by {} that did not match the key {} and"
                " mask {}".format(
                    self._label, self._virtual_key, self._mask))))
        provenance_items.append(ProvenanceDataItem(
            self._add_name(names, "incorrect_packets"),
            provenance_data[self._PROVENANCE_ITEMS.INCORRECT_PACKETS],
            report=provenance_data[
                self._PROVENANCE_ITEMS.INCORRECT_PACKETS] > 0,
            message=(
                "SDP Packets were received by {} that were not correct".format(
                    self._label))))
        provenance_items.append(ProvenanceDataItem(
            self._add_name(names, "late_packets"),
            provenance_data[self._PROVENANCE_ITEMS.LATE_PACKETS],
            report=provenance_data[self._PROVENANCE_ITEMS.LATE_PACKETS] > 0,
            message=(
                "SDP Packets were received by {} that were too late to be"
                " transmitted in the simulation".format(self._label))))

        return provenance_items

    def __repr__(self):
        return self._label
